/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
* Module Name:
*   cm.c
*   
* Abstract:
*   This file implements the Connection Manager Module.
*                                                                             
* Authors:
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date       Name  Description 
*   12-Oct-94  TLR   Created.
*   14-Sep-95  TLR   Updated CM-DRV interface for pool management and link
*                    events.
*   19-Sep-95  TLR   Added VC Aging.
*
* Description:
*/

#ifdef _DEBUG
#include <assert.h>
#else
#define assert(exp)
#endif

/* Include general headers. */
#include "g_types.h"
#include "codes.h"
#include "atm.h"
#include "utl.h"
#include "utl_os.h"
#include "system.h"
#include "svc_info.h"
#include "g_event.h"

/* Include headers for CM Interfaces. */
#include "cm.h"
#include "cm_sap.h"
#include "cm_drv.h"
#include "cm_svc.h"

/* Include headers for the CM's servers. */
#include "map.h"
#include "map_data.h"

#define EMOD   MOD_CM
#define EINST  p_cm->p_text
#define EMASK  p_cm->event_mask


/****************************************************************************
 * Internal Data Structures
 ****************************************************************************/

/* SAP_RECORD
 *
 *   This structure describes a single Service Access Point (SAP).  An instance
 *   of this structure is created each time a client registers with the SAP
 *   interface to the CM.  It contains the following members:
 *
 *   p_next.................A reference pointer to another SAP_RECORD instance.
 *                          This pointer is placed as the first member of the
 *                          structure to meet the requirements of the
 *                          SIMPLE_LIST construct.  Groups of SAP_RECORDs can
 *                          be stored in SIMPLE_LISTs as a result.
 *
 *   p_text.................A pointer to a text constant that describes this
 *                          SAP instance.  The text is provided by the SAP
 *                          client during registration.  The purpose of the
 *                          text string is to aid in management and debugging.
 *                          A description of the SAP can be easily obtained
 *                          simply by looking at the structure.
 *
 *   p_sap_rcv..............The address of the SAP client's receive callback
 *                          function.
 *
 *   p_sap_notify...........The address of the SAP client's notification
 *                          callback function.
 *
 *   p_sap_connect..........The address of the SAP client's connect request
 *                          callback function.
 *
 *   cm_handle..............The handle of the CM instance that the SAP client
 *                          registered with.  The cm_handle references state
 *                          associated with the CM instance (one per adapter).
 *                          The handle is placed here so each call to the CM
 *                          SAP interface doesn't need to supply this handle
 *                          as well as the sap_handle.
 *
 *   drv_handle.............The handle of the Driver instance associated with
 *                          this CM instance.  This data is redundant with an
 *                          entry in the CM_INSTANCE block.  It is placed here
 *                          as an optimization.  This handle will be needed
 *                          during transmit operations which are in the fast
 *                          path.  A copy is placed here for quick reference
 *                          rather than requiring further indirection into the
 *                          CM_INSTANCE block.
 *
 *   The following members are statistics counters that keep track of data
 *   per SAP.
 *
 *   xmt_pkt_count..........The total number of transmitted packets since
 *                          the SAP was created.
 *
 *   rcv_pkt_count..........The total number of received packets since the
 *                          SAP was created.
 *
 *   out_call_count.........The total number of outgoing SVC connect requests
 *                          since the SAP was created.
 *
 *   accept_call_count......The total number of incoming SVC connect requests
 *                          accepted by this SAP since the SAP was created.
 *
 *   reject_call_count......The total number of incoming SVC connect requests
 *                          that were rejected by this SAP since the SAP was
 *                          created.
 */
typedef struct _sap_record
   {
   struct _sap_record        *p_next;
   char                      *p_text;
   HANDLE                     cm_handle;
   HANDLE                     os_handle;
   HANDLE                     svc_handle;
   HANDLE                     sap_context;
   SAP_RCV_CALLBACK           rcv_callback;
   SAP_VC_NOTIFY_CALLBACK     vc_notify_callback;
   SAP_CONNECT_CALLBACK       svc_connect_callback;
   SAP_LINK_STATUS_CALLBACK   link_status_callback;
   SAP_XMT_DONE_CALLBACK      xmt_done_callback;

   UINT32                     xmt_packet_count;
   UINT32                     rcv_packet_count;
   UINT32                     out_call_count;
   UINT32                     accept_call_count;
   UINT32                     reject_call_count;
   } SAP_RECORD;

/* SAP_LIST
 *
 *   This type represents a simple list of SAP_RECORD entries.
 */
SIMPLE_LIST (SAP_RECORD, SAP_LIST);

typedef struct _pool_record
   {
   struct _pool_record  *p_next;
   HANDLE                context;
   POOL_TYPE             pool_type;
   UINT32                buffer_size;
   UINT32                depth;
   UINT32                assignment_mask;
   } POOL_RECORD;

SIMPLE_LIST (POOL_RECORD, POOL_LIST);

/* CM_INSTANCE
 *
 *   This structure describes a single CM instance.  When an instance is
 *   created (cm_create), a CM_INSTANCE structures is allocated and
 *   initialized.  The cm_handle returned to the caller of cm_create is
 *   the address of the allocated structure.
 *
 *   CM_INSTANCE contains the following members:
 *
 *   p_text.................A pointer to a text constant that describes this
 *                          CM instance.  This constant may be used for debug
 *                          and management.  The CM instance may easily be
 *                          identified by looking at the contents of this
 *                          string.
 *
 *   drv_handle.............The handle of the device driver instance that is
 *                          associated with this CM instance.  All calls to the
 *                          device driver shall include this handle.
 *
 *   vc_list................Handle of the VC list for this CM instance.  Note
 *                          that this list is shared with the LAN Emulation
 *                          Client module if such a module exists.
 *
 *   atm_list...............Handle of the ATM address list for this CM
 *                          instance.  Like the vc_list, this list is also
 *                          shared with the LEC.
 *
 *   sig_handle.............The handle of the signaling module instance that is
 *                          associated with this CM instance.  All calls to the
 *                          signaling module shall include this handle.
 *
 *   os_handle..............The handle of the operating system specific data
 *                          block associated with this CM instance.  All calls
 *                          to operating system utilities shall include this
 *                          handle.
 *
 *   sap_list...............A simple list of the SAP records for all currently
 *                          registered SAP clients.  This list is used to keep
 *                          track of SAP clients for management and for 
 *                          operations which need to access all SAP records
 *                          (destroy instance for example).
 *
 *   The following members are statistics counters that keep track of data
 *   per CM instance.
 *
 *   out_call_count.........The total number of outgoing call requests made
 *                          since creation of the instance.
 *
 *   in_call_count..........The total number of incoming calls received since
 *                          creation of the instance.
 *
 *   reject_no_saps_count...The total number of incoming calls that were
 *                          rejected due to no SAPs being bound to the call's
 *                          sap_id.
 *
 *   reject_all_saps_count..The total number of incoming calls that were
 *                          rejected due to rejection by all SAPs bound to the
 *                          call's sap_id.
 *
 *   rcv_unknown_count......The total number of packets received for which the
 *                          VC was not open.
 */
typedef struct
   {
   char                 *p_text;
   UINT32                event_mask;
   HANDLE                drv_handle;
   HANDLE                vc_list;
   HANDLE                atm_list;
   HANDLE                svc_handle;
   HANDLE                os_handle;
   SAP_LIST              sap_list;
   POOL_LIST             xmt_pool_list;
   POOL_LIST             rcv_pool_list;
   DRV_CM_VC_SETUP       drv_cm_vc_setup;
   DRV_CM_VC_TEARDOWN    drv_cm_vc_teardown;
   DRV_CM_XMT            drv_cm_xmt;
   ATM_LINK_STATE        link_state;
   BOOLEAN               link_down_flag;
   BOOLEAN               phy_resetting;
   HANDLE                hysteresis_timer;
   HANDLE                age_timer;
   CALL_ID               next_user_id;
   AAL_DATA              default_aal_data;

   LINK_CALLBACK         mac_linkage_update;
   LINK_CALLBACK         mac_linkage_remove;
   HANDLE                mac_context;

   UINT32                out_call_count;
   UINT32                in_call_count;
   UINT32                reject_no_saps_count;
   UINT32                reject_all_saps_count;
   UINT32                rcv_unknown_count;
   } CM_INSTANCE;

#define AGE_TIME_INTERVAL  10       /* Ten Seconds */

/*++
* =============================
* = VC_ID Manipulation Macros =
* =============================
*
* Overview:
*   VC_IDs are single 32-bit values that combine the VPI and VCI of a virtual
*   channel.  They are combined into a single value to facilitate quick lookup
*   in the vc_list.
*
*   These macros convert vpi/vci to vc_id and extract vpi and vci from a vc_id.
*
* Usage:
*   vc_id = VC_ID_BUILD (vpi, vci);
*   vpi   = VPI_GET (vc_id);
*   vci   = VCI_GET (vc_id);
*              
--*/
#define VC_ID_BUILD(vpi,vci) (((UINT32) vpi << 16) | (UINT32) vci)
#define VPI_GET(vc_id)       ((UINT8)  ((vc_id >> 16) & 0xff))
#define VCI_GET(vc_id)       ((UINT16) ( vc_id        & 0xffff))

/*++
* =======================
* = hysteresis_callback =
* =======================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Overview:
*   This is the callback invoked by the link_status hysteresis timer.  The
*   timer is set when the device driver indicates that the link status has
*   transitioned to LINK_DOWN.  In this case, the link_down_flag is set but
*   the link state is not changed.  When the timer expires, this callback
*   checks the link_down_flag.  If the flag is still set, the link state is
*   changed to LINK_DOWN and the SAP clients are notified.
*
* Arguments:
*   cm_handle   - (IN) Handle of the CM instance detecting the link state
*                      change.
*              
* Returns:
*   None
*   
* Preconditions:
*   None
*
* Postconditions:
*   If the link_down_flag is set, the link state is changed to LINK_DOWN.
*
--*/
static void hysteresis_callback (HANDLE cm_handle)
   {
   CM_INSTANCE  *p_cm;
   SAP_RECORD   *p_sap;

   p_cm = (CM_INSTANCE *) cm_handle;

   if (p_cm->link_down_flag)
      {
      p_cm->link_state     = LINK_DOWN;
      p_cm->link_down_flag = FALSE;

      /* Notify the SAP clients of the change in status. */

      utl_list_traverse (p_cm->sap_list, p_sap)
         {
         if (p_sap->link_status_callback != NULL)
            p_sap->link_status_callback (p_sap->sap_context, p_cm->link_state);
         }
      }
   }

static BOOLEAN svc_teardown (HANDLE, HANDLE);
static void vc_linkage_update (HANDLE conn_handle);

static BOOLEAN svc_age_advance (HANDLE conn_handle,
                                HANDLE cm_handle)
   {
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;

   p_cm = (CM_INSTANCE   *) cm_handle;
   p_vc = (MAP_VC_RECORD *) conn_handle;

   if ((p_vc->age_limit != 0) &&
       ((p_vc->vc_flags & (VC_FLAG_PVC | VC_FLAG_RVC)) == 0) &&
       (VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPEN))
      {
      p_vc->idle_time += AGE_TIME_INTERVAL;

      if (p_vc->idle_time >= p_vc->age_limit)
         {
         svc_teardown (conn_handle, cm_handle);
         }
      }
   return TRUE;
   }

static void age_timer_handler (HANDLE cm_handle)
   {
   CM_INSTANCE  *p_cm;

   p_cm = (CM_INSTANCE *) cm_handle;

   os_timer_set (p_cm->age_timer, AGE_TIME_INTERVAL * 1000);

   map_list_traverse (p_cm->vc_list, &svc_age_advance, cm_handle);
   }

typedef struct
   {
   CALL_ID  call_user_id;
   BOOLEAN  found_flag;
   HANDLE   conn_handle;
   } UI_FIND_BLOCK;

/*++
* =========================
* = user_id_find_callback =
* =========================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   This callback routine is invoked during traversal of the vc_list mapping.
*   Refer to "call_user_id_find" to see where it is used.
*
*   This routine checks to see if the current connection element is the one
*   with a call_user_id that matches the requested call_user_id.  If so, it
*   returns FALSE (to stop the traversal) and places the connection handle in
*   the parameter block to be used by "call_user_id_find".
*
* Arguments:
*   conn_handle - (IN) Handle of connection record for this iteration.
*   param       - (IN) Parameter handle for the UI_FIND_BLOCK structure.
*              
* Returns:
*   Boolean value indicating whether the traversal should continue or halt.
*   
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
static BOOLEAN user_id_find_callback (HANDLE conn_handle,
                                      HANDLE param)
   {
   MAP_VC_RECORD  *p_vc;
   UI_FIND_BLOCK  *p_ui;

   /* Convert the input handles to structure pointers. */

   p_vc = (MAP_VC_RECORD *) conn_handle;
   p_ui = (UI_FIND_BLOCK *) param;

   /* Compare the call_user_ids in the param block and the connection block. */

   if (p_vc->call_user_id == p_ui->call_user_id)
      {
      /* If the call_user_ids match, set the found flag, copy the conn_handle
       * into the param block, and return FALSE to halt the traversal.
       */
      p_ui->found_flag  = TRUE;
      p_ui->conn_handle = conn_handle;
      return FALSE;
      }
   else
      {
      /* If the call_user_ids don't match, return TRUE to continue looking. */

      return TRUE;
      }
   }

/*++
* =====================
* = call_user_id_find =
* =====================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   Search the vc_list for a VC record using the call_user_id as a key instead
*   of the vc_id.  This is used to do lookups on VCs when the only
*   distinguishing known value is the call_user_id.
*
* Arguments:
*   cm_handle       - (IN)  Handle of the cm_instance doing the lookup.
*   call_user_id    - (IN)  Search key for the lookup.
*   p_conn_handle   - (OUT) Handle of connection record if a match is found.
*              
* Returns:
*   STATUS_K_SUCCESS    - A matching record was found.
*   STATUS_K_NOT_FOUNd  - No matching record was found.
*   
* Preconditions:
*   None
*
* Postconditions:
*   If status is STATUS_K_SUCCESS, the record referenced by *p_conn_handle
*   contains a call_user_id matching the indicated call_user_id.
--*/
static STATUS call_user_id_find (HANDLE    cm_handle,
                                 CALL_ID   call_user_id,
                                 HANDLE   *p_conn_handle)
   {
   CM_INSTANCE     *p_cm;
   UI_FIND_BLOCK   user_id_find_block;

   /* Convert the cm_handle into a structure pointer. */

   p_cm = (CM_INSTANCE *) cm_handle;

   /* Initialize the find_block structure for the traversal. */

   user_id_find_block.call_user_id = call_user_id;
   user_id_find_block.found_flag   = FALSE;

   /* Traverse the VC list checking each entry using the
    * "user_id_find_callback" routine.
    */
   map_list_traverse (p_cm->vc_list,
                      &user_id_find_callback,
                      (HANDLE) &user_id_find_block);

   /* Test the found_flag member of the find_block structure. */

   if (user_id_find_block.found_flag)
      {
      *p_conn_handle = user_id_find_block.conn_handle;
      return STATUS_K_SUCCESS;
      }
   else
      return STATUS_K_NOT_FOUND;
   }

/*++
* ================
* = svc_teardown =
* ================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   Begin the process of tearing down an SVC.  Check the svc_status to see if
*   the SAP client needs to be notified of the teardown and check the
*   phy_status to see if the lower device driver needs to be told to teardown
*   the circuit.
*
*   This routine is meant to be used as a map_traverse handler.  It may also
*   be called directly to teardown an SVC.  When used as a map_traverse
*   handler, the traversal is done to tear down all SVCs in the vc_list.
*
* Arguments:
*   conn_handle - (IN) Connection handle of VC to be torn down.
*   cm_handle   - (IN) Handle of the CM instance being used.
*              
* Returns:
*   TRUE    - This boolean value is always returned.  This is to ensure that
*             the map_list_traverse routine will always continue to the end
*             of the vc_list.  This value should be ignored by direct callers.
*   
* Preconditions:
*   None
*
* Postconditions:
*   If the VC record contains an SVC, the svc_status and phy_status of the VC
*   are SVC_CLOSING and PHY_CLOSING respectively.
*
* Issues:
*   This routine tests the svc_status and phy_status values to determine if
*   the SAP client should be notified of the teardown and whether the lower
*   device driver should be instructed to close the VC.  It is cricital that
*   these requests be made only once for each SVC that is torn down.  It is
*   possible that a SAP client will request an SVC teardown at the same time
*   the network requests the teardown of the same SVC.  In this case, there
*   MUST be only one notification call to the SAP client and one call to the
*   lower device driver to close the VC.  There is a critcal section shown in
*   this routine.  It is assumed that mutual exclusion is ensured by external
*   means.
--*/
static BOOLEAN svc_teardown (HANDLE conn_handle,
                             HANDLE cm_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   BOOLEAN         svc_flag;
   BOOLEAN         phy_flag;
   RELEASE_INFO    rel_info;

   p_cm = (CM_INSTANCE *)   cm_handle;
   p_vc = (MAP_VC_RECORD *) conn_handle;

   if ((p_vc->vc_flags & (VC_FLAG_PVC | VC_FLAG_RVC)) == 0)
      {
      svc_flag = FALSE;
      phy_flag = FALSE;

      /* CRITICAL_SECTION_ENTER */
      if (VC_SVC_STATE (p_vc->vc_flags) != VC_STATE_CLOSING)
         {
         VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);
         svc_flag = TRUE;
         }
      if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN)
         {
         VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);
         phy_flag = TRUE;
         }
      /* CRITICAL_SECTION_EXIT */

      vc_linkage_update (conn_handle);

      if (svc_flag)
         {
         p_sap = (SAP_RECORD *) p_vc->sap_handle;
         if ((p_sap->vc_notify_callback != NULL) &&
             (p_vc->vc_flags & VC_FLAG_NOTIFICATION_REQUIRED))
            {
            p_sap->vc_notify_callback (p_vc->conn_context,
                                       (HANDLE) p_vc,
                                       CONN_RELEASE,
                                       (UINT16) CAUSE_NORMAL,
                                       (UINT32) NULL,
                                       VC_CALLING_PARTY (p_vc->vc_flags));

            p_vc->vc_flags &= ~VC_FLAG_NOTIFICATION_REQUIRED;
            }

         /* Send a release to the signaling stack only if the
          * signaling function is up.
          */
         if (p_cm->link_state == LINK_SIG_UP)
            {
            svc_release_info_make (CAUSE_NORMAL, &rel_info);
            svc_cm_rel_req (p_cm->svc_handle,
                            p_vc->call_user_id,
                            p_vc->call_provider_id,
                            &rel_info);
            }
         }

      if (phy_flag)
         {
         p_cm->drv_cm_vc_teardown (p_cm->drv_handle,
                                   VPI_GET (p_vc->vc_id),
                                   VCI_GET (p_vc->vc_id),
                                   p_vc->xmt_pool_context,
                                   p_vc->rcv_pool_context);
         }
      }

   return TRUE;
   }

/*++
* ==================
* = atm_is_greater =
* ==================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   Compare two ATM addresses to determine if the first is numerically greater than
*   the second.
*
* Arguments:
*   a1   - (IN) First ATM Address
*   a2   - (IN) Second ATM Address
*
* Returns:
*   TRUE    - If a1 is numerically greater than a2
*   FALSE   - Otherwise.
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
static BOOLEAN atm_is_greater (ADDR_ATM a1,
                               ADDR_ATM a2)
   {
   UINT8    *p_addr1;
   UINT8    *p_addr2;
   int       i;

   p_addr1 = (UINT8 *) &a1;
   p_addr2 = (UINT8 *) &a2;

   for (i = 0; i < sizeof (ADDR_ATM); i++)
      {
      if (p_addr1[i] > p_addr2[i])
         return TRUE;
      if (p_addr1[i] < p_addr2[i])
         return FALSE;
      }

   return FALSE;
   }

/*++
* ===========================
* = atm_active_vc_calculate =
* ===========================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   Complete the linkage from the indicated ATM address record to a VC record
*   if possible.  If there is more than one VC associated with the ATM address,
*   choose the one with the numerically lower caller_atm_address.  (Refer to
*   section 8.1.11 of the LAN Emulation Spec version 1.0).
*
* Arguments:
*   atm_handle  - (IN) Handle of the ATM address record to be updated.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The ATM Address's atm_status is up-to-date.
*
*   If there are zero vc_records associated with the atm_record, the active
*   VC handle is NULL.
--*/
void atm_active_vc_calculate (HANDLE  atm_handle)
   {
   MAP_ATM_RECORD   *p_atm;
   MAP_VC_RECORD    *p_vc;
   MAP_VC_RECORD    *p_active;
   HANDLE_RECORD    *p_link;

   p_atm = (MAP_ATM_RECORD *) atm_handle;

   p_atm->active_vc_link = (HANDLE) NULL;

   utl_list_traverse (p_atm->vc_link_list, p_link)
      {
      p_vc = (MAP_VC_RECORD *) p_link->handle;

      /* First check to see if the VC is open.  If it is not, it cannot be
       * the active VC.
       */
      if ((VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPEN) &&
          (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN))
         {
         if (p_atm->active_vc_link == (HANDLE) NULL)
            {
            /* If this is the first vc_link in the list, no comparison is
             * required, simply assume it is the active link.
             */
            p_atm->active_vc_link = p_vc;
            }
         else
            {
            /* If this is not the first vc_link in the list, compare with the
             * current active link.  If the active link has a numerically
             * greater calling party address, set the active link to this link.
             */
            p_active = (MAP_VC_RECORD *) p_atm->active_vc_link;
            if (atm_is_greater (p_active->calling_party, p_vc->calling_party))
               p_atm->active_vc_link = p_vc;
            }
         }
      }

   /* Calculate the new atm_status for this record.  If there is an active
    * link, the status becomes ATM_VALID.  If there is no active link but there
    * is at least one link in the vc_link_list, the status becomes ATM_PENDING
    * or ATM_FAIL depending on the status of the VCs in the list.  If there
    * are no links in the vc_link_list, the status becomes ATM_NONE.
    */
   if      (p_atm->active_vc_link != (HANDLE) NULL)
      p_atm->atm_status = ATM_VALID;
   else if (p_atm->vc_link_list.size > 0)
      {
      p_atm->atm_status = ATM_FAIL;
      utl_list_traverse (p_atm->vc_link_list, p_link)
         {
         p_vc = (MAP_VC_RECORD *) p_link->handle;
         if (VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPENING)
            p_atm->atm_status = ATM_PENDING;
         }
      }
   else
      p_atm->atm_status = ATM_NONE;
   }

/*++
* =====================
* = vc_linkage_remove =
* =====================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   This routine is called in preparation for deleting a VC record.  It
*   removes the back link to the ATM address record, the forward link from
*   the ATM address record to this VC record, and notifies all MAC address
*   records (for LAN Emulation) of the change in status.  If the associated
*   ATM address record is left with no links to VC or MAC records, the ATM
*   address record is removed from the ATM list and is deallocated.
*
* Arguments:
*   conn_handle   - (IN) Handle of the connection being unlinked.
*              
* Returns:
*  None
*   
* Preconditions:
*  None
*
* Postconditions:
*  The VC record referred to by the conn_handle is not linked to an ATM address
*  record and no ATM address records or MAC address records contain links to
*  the VC record.
--*/
static void vc_linkage_remove (HANDLE conn_handle)
   {
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;
   MAP_ATM_RECORD *p_atm;
   HANDLE_RECORD  *p_link;

   p_vc  = (MAP_VC_RECORD  *) conn_handle;
   p_atm = (MAP_ATM_RECORD *) p_vc->atm_link;
   p_cm  = (CM_INSTANCE    *) p_vc->cm_handle;

   if (p_atm != NULL)
      {
      /* Remove the back link from the VC record to the ATM record. */

      p_vc->atm_link = (HANDLE) NULL;

      /* Remove the forward link from the ATM VC link list. */

      utl_list_traverse (p_atm->vc_link_list, p_link)
         {
         if (p_link->handle == conn_handle)
            {
            utl_list_delete (p_atm->vc_link_list, p_link);
            os_mem_dealloc (p_cm->os_handle, p_link);
            break;
            }
         }

      /* Calculate a new active VC for the ATM address record.  If the last
       * VC link was just removed from the vc_link_list, the new active VC
       * link shall be NULL.
       */
      atm_active_vc_calculate ((HANDLE) p_atm);

      if (p_atm->esi_link_list.size > 0)
         {
         /* If there are back links to MAC address records, traverse the
          * list of back links and update each one's linkage.
          */
         utl_list_traverse (p_atm->esi_link_list, p_link)
            {
            if (p_cm->mac_linkage_update != NULL)
               p_cm->mac_linkage_update (p_cm->mac_context, p_link->handle);
            }
         }
      else
         {
         /* If there are no back links AND the last VC for the ATM address
          * has been removed, the ATM address record must be removed from
          * the atm_list and deallocated.
          */
         if (p_atm->vc_link_list.size == 0)
            {
            map_element_delete (p_cm->atm_list, p_atm);
            os_mem_dealloc (p_cm->os_handle, p_atm);
            }
         }
      }
   }

/*++
* =====================
* = vc_linkage_update =
* =====================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Overview:
*   Update the linkage to and from a VC record.  This is done each time a VC
*   changes status or is added to the vc_list.
*
* Arguments:
*   conn_handle - (IN) Handle that references the VC record.
*              
* Returns:
*   None.
*   
* Preconditions:
*   There is an ATM address record linked to this VC record.
*
* Postconditions:
*   The linkage from the ATM address record is current and contains the correct
*   active link.
--*/
static void vc_linkage_update (HANDLE conn_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   MAP_ATM_RECORD *p_atm;
   HANDLE_RECORD  *p_link;
   HANDLE          saved_link;
   ATM_STATUS      saved_atm_status;

   p_vc  = (MAP_VC_RECORD  *) conn_handle;
   p_sap = (SAP_RECORD     *) p_vc->sap_handle;
   p_atm = (MAP_ATM_RECORD *) p_vc->atm_link;
   p_cm  = (CM_INSTANCE    *) p_sap->cm_handle;

   /* If this VC is not an SVC, return without doing anything. */

   if (p_vc->vc_flags & (VC_FLAG_PVC | VC_FLAG_RVC))
      return;

   if (p_atm != NULL)
      {
      /* Recalculate the ATM address's active VC link. */
      saved_link       = p_atm->active_vc_link;
      saved_atm_status = p_atm->atm_status;
      atm_active_vc_calculate ((HANDLE) p_atm);

      /* If the active VC link changed or the ATM status has changed,
       * notification must be made along the MAC address back links.
       */
      if ((saved_link       != p_atm->active_vc_link) ||
          (saved_atm_status != p_atm->atm_status))
         {
         utl_list_traverse (p_atm->esi_link_list, p_link)
            {
            if (p_cm->mac_linkage_update != NULL)
               p_cm->mac_linkage_update (p_cm->mac_context, p_link->handle);
            }
         }
      }
   else
      {
      /* If this code block is executed, the precondition requiring that there
       * be an ATM address record linked to the VC record was violated.
       */
      assert (FALSE);
      }
   }

/*++
* ==================
* = vc_record_make =
* ==================
*
* Status:    IMPLEMENTATION COMPLETE
*
* Overview:
*   Allocate a new VC record and associate it with the proper ATM address
*   record.  If no ATM address record exists with the indicated ATM address,
*   a new ATM address record is allocated and added to the atm_list.
*
* Arguments:
*   cm_handle     - (IN)  Handle of this CM instance.
*   atm_present   - (IN)  Boolean flag.  If true, an ATM address is supplied
*                         in the "addr_atm" argument.  Otherwise there is no
*                         ATM address supplied.
*   addr_atm      - (IN)  ATM address associated with this new VC.
*   p_conn_handle - (OUT) New connection handle with proper linkages set up.
*              
* Returns:
*   STATUS_K_SUCCESS    - Successful completion - See postconditions
*   STATUS_K_RESOURCES  - Insufficient resources to complete operation.
*   
* Preconditions:
*
* Postconditions:
*   The ATM address record and VC record are properly linked together.
*
*   Both the ATM address record and the VC record are members of the atm_list
*   and the vc_list respectively.
*
*   The phy_status and svc_status of the VC record are PHY_NONE and SVC_NONE
*   respectively.
--*/
static STATUS vc_record_make (HANDLE    cm_handle,
                              BOOLEAN   atm_present,
                              ADDR_ATM  addr_atm,
                              HANDLE   *p_conn_handle)
   {
   CM_INSTANCE      *p_cm;
   MAP_VC_RECORD    *p_vc;
   MAP_ATM_RECORD   *p_atm;
   HANDLE            atm_handle;
   HANDLE_RECORD    *p_link;
   STATUS            status;

   p_cm = (CM_INSTANCE *) cm_handle;

   /* Allocate memory for a VC record.  Return error status if allocation is
    * unsuccessful.
    */
   p_vc = (MAP_VC_RECORD *) os_mem_alloc (p_cm->os_handle,
                                          sizeof (MAP_VC_RECORD));
   if (p_vc == NULL)
      return STATUS_K_RESOURCES;

   /* Initialize some of the fields of the new record. */

   p_vc->vc_flags = VC_FLAGS_INIT;

   p_vc->cm_handle         = cm_handle;
   p_vc->idle_time         = 0;
   p_vc->age_limit         = 0;
   p_vc->xmt_packet_count  = 0;
   p_vc->rcv_packet_count  = 0;
   p_vc->call_provider_id  = 0;
   p_vc->call_user_id      = p_cm->next_user_id;
   p_cm->next_user_id     += 1;
   if (p_cm->next_user_id & 0xffff0000)
      p_cm->next_user_id = 0;

   if (atm_present)
      {
      /* Create a link element to be used to link the VC to the ATM address. */

      p_link = (HANDLE_RECORD *) os_mem_alloc (p_cm->os_handle,
                                               sizeof (HANDLE_RECORD));
      if (p_link == NULL)
         {
         os_mem_dealloc (p_cm->os_handle, p_vc);
         return STATUS_K_RESOURCES;
         }

      /* Search the ATM address list for a record containing the indicated
       * ATM address.
       */
      status = map_key_search (p_cm->atm_list, &addr_atm, &atm_handle);

      /* If no matching ATM address record was found, we must create a new
       * one.
       */
      if (status != STATUS_K_SUCCESS)
         {
         /* Allocate a new ATM address record. */

         p_atm = (MAP_ATM_RECORD *) os_mem_alloc (p_cm->os_handle,
                                                  sizeof (MAP_ATM_RECORD));
      
         /* If the allocation failed, deallocate the new VC record and return
          * with error status.
          */
         if (p_atm == NULL)
            {
            os_mem_dealloc (p_cm->os_handle, p_vc);
            os_mem_dealloc (p_cm->os_handle, p_link);
            return STATUS_K_RESOURCES;
            }

         /* Initialize the contents of the ATM address record. */

         ATM_COPY (addr_atm, p_atm->addr);
         p_atm->atm_status     = ATM_PENDING;
         p_atm->active_vc_link = (HANDLE) NULL;

         utl_list_init (p_atm->vc_link_list);
         utl_list_init (p_atm->esi_link_list);

         /* Insert the new ATM address record into the atm_list. */

         map_element_insert (p_cm->atm_list, (HANDLE) p_atm);
         }
      else
         {
         p_atm = (MAP_ATM_RECORD *) atm_handle;
         }

      /* Add a forward link from the ATM address record to the VC record.
       * Note that it is not necessary to recalculate the active link because
       * this new VC is not yet open and therefore cannot be the active VC for
       * the ATM address.
       */
      p_link->handle = (HANDLE) p_vc;
      utl_list_add (p_atm->vc_link_list, p_link);
      }
   else
      {
      /* If there is no ATM address to be associated with this VC. */

      p_atm = NULL;
      }

   /* Update the back link from the VC record to the ATM address record. */

   p_vc->atm_link = (HANDLE) p_atm;

   /* Insert the new VC record into the VC list. */

   map_element_insert (p_cm->vc_list, (HANDLE) p_vc);

   /* Return the connection handle with success status. */

   *p_conn_handle = (HANDLE) p_vc;
   return STATUS_K_SUCCESS;
   }

/*++
* ================
* = cm_qos_check =
* ================
*
* Status:      TEST VERSION COMPLETE
*
* Overview:
*
* Arguments:
*              
* Returns:
*   
* Preconditions:
*
* Postconditions:
--*/
static STATUS cm_qos_check (HANDLE cm_handle,
                            QOS    requested_qos)
   {
   return STATUS_K_SUCCESS;
   }

/*++
* ==================
* = cm_pool_assign =
* ==================
*
* Status:      TEST VERSION COMPLETE
*
* Overview:
*
* Arguments:
*              
* Returns:
*   
* Preconditions:
*
* Postconditions:
--*/
static void cm_pool_assign (HANDLE     sap_handle,
                            QOS       *p_requested_qos,
                            HANDLE    *p_xmt_pool_context,
                            HANDLE    *p_rcv_pool_context)
   {
   /* This is a temporary solution:  Assign first pool. */
   CM_INSTANCE  *p_cm;
   SAP_RECORD   *p_sap;

   p_sap = (SAP_RECORD  *) sap_handle;
   p_cm  = (CM_INSTANCE *) p_sap->cm_handle;

   *p_xmt_pool_context = ((POOL_RECORD *) p_cm->xmt_pool_list.p_root)->context;
   *p_rcv_pool_context = ((POOL_RECORD *) p_cm->rcv_pool_list.p_root)->context;
   }

/****************************************************************************
 * Implementations of General Functions (from cm.h)
 ****************************************************************************/

/*++
* =============
* = cm_create =
* =============
*
* Status:
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_create(HANDLE  os_handle,
                 char   *p_text,
                 HANDLE *p_cm_handle)
   {
   CM_INSTANCE    *p_cm;
   STATUS          status;

   p_cm = (CM_INSTANCE*) os_mem_alloc (os_handle, sizeof (CM_INSTANCE));
   if (p_cm == NULL)
      return STATUS_K_RESOURCES;

   status = os_timer_alloc (os_handle,
                            &hysteresis_callback,
                            (HANDLE) p_cm,
                            &p_cm->hysteresis_timer);
   if (status != STATUS_K_SUCCESS)
      {
      os_mem_dealloc (os_handle, p_cm);
      return STATUS_K_RESOURCES;
      }

   status = os_timer_alloc (os_handle,
                            &age_timer_handler,
                            (HANDLE) p_cm,
                            &p_cm->age_timer);
   if (status != STATUS_K_SUCCESS)
      {
      os_timer_dealloc (p_cm->age_timer);
      os_mem_dealloc (os_handle, p_cm);
      return STATUS_K_RESOURCES;
      }

   os_timer_set (p_cm->age_timer, AGE_TIME_INTERVAL * 1000);

   p_cm->p_text      = p_text;
   p_cm->drv_handle  = NULL;
   p_cm->svc_handle  = NULL;
   p_cm->os_handle   = os_handle;

   map_list_alloc (os_handle, sizeof (UINT32),   &p_cm->vc_list);
   map_list_alloc (os_handle, sizeof (ADDR_ATM), &p_cm->atm_list);

   p_cm->out_call_count        = 0;
   p_cm->in_call_count         = 0;
   p_cm->reject_no_saps_count  = 0;
   p_cm->reject_all_saps_count = 0;
   p_cm->rcv_unknown_count     = 0;

   p_cm->next_user_id          = 1;

   p_cm->drv_cm_vc_setup    = NULL;
   p_cm->drv_cm_vc_teardown = NULL;
   p_cm->drv_cm_xmt         = NULL;

   p_cm->mac_linkage_update = NULL;
   p_cm->mac_linkage_remove = NULL;

   p_cm->link_state     = LINK_DOWN;
   p_cm->link_down_flag = FALSE;
   p_cm->phy_resetting  = FALSE;

   p_cm->default_aal_data.aal5_data.cpcs_uu = 0;
   p_cm->default_aal_data.aal5_data.clp     = 0;

   utl_list_init (p_cm->sap_list);

   *p_cm_handle = (HANDLE) p_cm;

   return STATUS_K_SUCCESS;
   }

/*++
* =================
* = cm_drv_config =
* =================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_drv_config(HANDLE  cm_handle,
                   HANDLE  drv_handle)
   {
   CM_INSTANCE    *p_cm;

   p_cm = (CM_INSTANCE*) cm_handle;

   p_cm->drv_handle = drv_handle;
   }


/*++
* =================
* = cm_svc_config =
* =================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_config(HANDLE  cm_handle,
                   HANDLE  svc_handle)
   {
   CM_INSTANCE    *p_cm;

   p_cm = (CM_INSTANCE*) cm_handle;

   p_cm->svc_handle = svc_handle;
   }


/*++
* ==============
* = cm_map_get =
* ==============
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_map_get(HANDLE         cm_handle,
                LINK_CALLBACK  link_update_callback,
                LINK_CALLBACK  link_remove_callback,
                HANDLE         callback_context,
                HANDLE        *p_vc_list,
                HANDLE        *p_atm_list)
   {
   CM_INSTANCE  *p_cm;

   p_cm = (CM_INSTANCE*) cm_handle;

   p_cm->mac_linkage_update = link_update_callback;
   p_cm->mac_linkage_remove = link_remove_callback;
   p_cm->mac_context        = callback_context;

   *p_vc_list  = p_cm->vc_list;
   *p_atm_list = p_cm->atm_list;
   }

static BOOLEAN vc_destroy_callback (HANDLE conn_handle,
                                    HANDLE os_handle)
   {
   vc_linkage_remove (conn_handle);
   os_mem_dealloc (os_handle, (MAP_VC_RECORD *) conn_handle);
   return TRUE;
   }

static BOOLEAN atm_destroy_callback (HANDLE atm_handle,
                                     HANDLE os_handle)
   {
   os_mem_dealloc (os_handle, (MAP_ATM_RECORD *) atm_handle);
   return TRUE;
   }


/*++
* ==============
* = cm_destroy =
* ==============
*
* Status:
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_destroy(HANDLE  cm_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;

   /* Do nothing if the handle is already NULL. */

   if (cm_handle == NULL)
      return;
         
   p_cm = (CM_INSTANCE *) cm_handle;

   /* Destroy the mapping lists.  Note that no linkage update is performed.
    * It is assumed that a destroy is a major and far-reaching event and that
    * the other components including the lower device driver will be destroyed
    * and/or reset.
    */
   map_list_dealloc (p_cm->vc_list,  vc_destroy_callback,  p_cm->os_handle);
   map_list_dealloc (p_cm->atm_list, atm_destroy_callback, p_cm->os_handle);

   /* Delete the SAP records. */

   while (p_cm->sap_list.p_root != NULL)
      {
      p_sap = p_cm->sap_list.p_root;
      utl_list_delete (p_cm->sap_list, p_sap);
      os_mem_dealloc (p_cm->os_handle, p_sap);
      }

   /* Deallocate the hysteresis timer. */

   os_timer_dealloc (p_cm->hysteresis_timer);

   /* Deallocate the cm_instance block. */

   os_mem_dealloc (p_cm->os_handle, p_cm);
   }

/*++
* =====================
* = cm_event_mask_set =
* =====================
*
* Status:
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_event_mask_set (HANDLE   cm_handle,
                        UINT32   event_mask)
   {
   CM_INSTANCE    *p_cm;

   p_cm = (CM_INSTANCE *) cm_handle;

   p_cm->event_mask = event_mask;
   }

static BOOLEAN cm_open_pvc_only (HANDLE conn_handle,
                                 HANDLE cm_handle)
   {
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;

   p_cm = (CM_INSTANCE   *) cm_handle;
   p_vc = (MAP_VC_RECORD *) conn_handle;

   if (p_vc->vc_flags & VC_FLAG_PVC)
      {
      if (p_cm->drv_handle != NULL)
         {
         VC_PHY_STATE_SET(p_vc->vc_flags, VC_STATE_OPENING);
         p_cm->drv_cm_vc_setup (p_cm->drv_handle,
                                VPI_GET (p_vc->vc_id),
                                VCI_GET (p_vc->vc_id),
                                p_vc->xmt_pool_context,
                                p_vc->rcv_pool_context,
                                &p_vc->requested_qos);
         }
      }

   return TRUE;
   }

static void cm_pvc_reestablish (HANDLE cm_handle)
   {
   CM_INSTANCE    *p_cm;

   p_cm = (CM_INSTANCE *) cm_handle;

   map_list_traverse (p_cm->vc_list, cm_open_pvc_only, cm_handle);
   }

/*++
* =================
* = cm_link_event =
* =================
*
* Status:
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_link_event (HANDLE           cm_handle,
                    ATM_LINK_EVENT   link_event)
   {
   CM_INSTANCE       *p_cm;
   SAP_RECORD        *p_sap;
   ATM_LINK_STATE     new_state;

   /* Convert the cm_handle to a pointer to a CM_INSTANCE block. */

   p_cm = (CM_INSTANCE *) cm_handle;

/* Define a macro for combining a state value with an event value into a single
 * integer.  This is used to define state transitions (i.e. a combination of a
 * previous state and an event).
 */
#define COMBINE(state,event) (((UINT8) state << 8) | (UINT8) event)

   /* If the link_down_flag is set and the event is LINK_EVENT_PHY_UP, the link
    * has recovered before the hysteresis time period has passed.  In this
    * case, the flag must be cleared and the timer cancelled.
    */
   if ((link_event == LINK_EVENT_PHY_UP) && (p_cm->link_down_flag))
      {
      p_cm->link_down_flag = FALSE;
      os_timer_cancel (p_cm->hysteresis_timer);
      }

   /* Handle all possible link transitions. */

   new_state = p_cm->link_state;

   switch (COMBINE (p_cm->link_state, link_event))
      {
      /* Handle the normal transition from LINK_DOWN to LINK_PHY_UP. */

      case COMBINE (LINK_DOWN, LINK_EVENT_PHY_UP):
         new_state = LINK_PHY_UP;
         break;

      /* Handle the normal transition from LINK_PHY_UP to LINK_LINE_UP. */

      case COMBINE (LINK_PHY_UP, LINK_EVENT_LINE_UP):
         new_state = LINK_LINE_UP;
         break;

      /* Handle the normal transition from LINK_LINE_UP to LINK_SIG_UP. */

      case COMBINE (LINK_LINE_UP, LINK_EVENT_SIG_UP):
         new_state = LINK_SIG_UP;
         break;

      /* Handle the PHY_DOWN event. */

      case COMBINE (LINK_PHY_UP,  LINK_EVENT_PHY_DOWN):
      case COMBINE (LINK_LINE_UP, LINK_EVENT_PHY_DOWN):
      case COMBINE (LINK_SIG_UP,  LINK_EVENT_PHY_DOWN):
         if (p_cm->link_down_flag == FALSE)
            {
            p_cm->link_down_flag = TRUE;

            /* Set hysteresis timer to proper time period.
             * If the timer expires and the link_down_flag is still set,
             * the link status will be updated to LINK_DOWN.
             */
            os_timer_set (p_cm->hysteresis_timer, 3000 /* msec */);
            }
         break;

      /* Handle the SIG_DOWN event.  In this case, the state machine
       * transitions to the LINK_PHY_UP state.  This requires that line-up
       * be performed again before signaling tries to come back up.
       */
      case COMBINE (LINK_SIG_UP, LINK_EVENT_SIG_DOWN):
         new_state = LINK_PHY_UP;
         break;

      /* Handle the RESET_START event. */

      case COMBINE (LINK_DOWN,    LINK_EVENT_RESET_START):
      case COMBINE (LINK_PHY_UP,  LINK_EVENT_RESET_START):
      case COMBINE (LINK_LINE_UP, LINK_EVENT_RESET_START):
      case COMBINE (LINK_SIG_UP,  LINK_EVENT_RESET_START):
         new_state = LINK_DOWN;
         p_cm->phy_resetting = TRUE;
         break;

      /* Handle the RESET_COMPLETE event. */

      case COMBINE (LINK_DOWN, LINK_EVENT_RESET_COMPLETE):
         p_cm->phy_resetting = FALSE;
         cm_pvc_reestablish (cm_handle);
         break;

      /* Handle the cases in which an error should be issued. */

      case COMBINE (LINK_DOWN, LINK_EVENT_LINE_UP):
      case COMBINE (LINK_DOWN, LINK_EVENT_SIG_UP):
      case COMBINE (LINK_DOWN, LINK_EVENT_SIG_DOWN):
      case COMBINE (LINK_PHY_UP, LINK_EVENT_SIG_UP):
      case COMBINE (LINK_PHY_UP, LINK_EVENT_SIG_DOWN):
      case COMBINE (LINK_SIG_UP, LINK_EVENT_RESET_COMPLETE):
      case COMBINE (LINK_LINE_UP, LINK_EVENT_RESET_COMPLETE):
      case COMBINE (LINK_PHY_UP, LINK_EVENT_RESET_COMPLETE):
      case COMBINE (LINK_SIG_UP, LINK_EVENT_PHY_UP):
      case COMBINE (LINK_SIG_UP, LINK_EVENT_LINE_UP):
      case COMBINE (LINK_LINE_UP, LINK_EVENT_SIG_DOWN):
      case COMBINE (LINK_LINE_UP, LINK_EVENT_PHY_UP):
         EVENT (EM_NERR, ("Link event (%d) received in unexpected state (%d).\n",
                          link_event, p_cm->link_state));
         break;

      /* The following cases require no action.  They have been left in
       * commented form to complete the list of {STATE,EVENT} pairs.
       *
      case COMBINE (LINK_DOWN,    LINK_EVENT_PHY_DOWN):
      case COMBINE (LINK_PHY_UP,  LINK_EVENT_PHY_UP):
      case COMBINE (LINK_LINE_UP, LINK_EVENT_LINE_UP):
      case COMBINE (LINK_SIG_UP,  LINK_EVENT_SIG_UP):
       */
      default:
         break;
      }

   /* Notify the SAP clients of the change in state. */

   if (new_state != p_cm->link_state)
      {
      p_cm->link_state = new_state;
      utl_list_traverse (p_cm->sap_list, p_sap)
         {
         if (p_sap->link_status_callback != NULL)
            p_sap->link_status_callback (p_sap->sap_context, p_cm->link_state);
         }
      }
   }

/****************************************************************************
 * SAP Interface Routines (cm_sap.h)
 ****************************************************************************/

/*++
* ===================
* = cm_sap_register =
* ===================
*
* Status:   IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_register (HANDLE                     cm_handle,
                        HANDLE                     sap_context,
                        SAP_RCV_CALLBACK           rcv_callback,
                        SAP_VC_NOTIFY_CALLBACK     vc_notify_callback,
                        SAP_CONNECT_CALLBACK       svc_connect_callback,
                        SAP_LINK_STATUS_CALLBACK   link_status_callback,
                        SAP_XMT_DONE_CALLBACK      xmt_done_callback,
                        char                      *p_sap_text,
                        HANDLE                    *p_sap_handle)
   {
   CM_INSTANCE *p_cm;
   SAP_RECORD  *p_sap;

   /* Convert the cm_handle to a reference to an instance record. */

   p_cm = (CM_INSTANCE*) cm_handle;

   /* Allocate a SAP record for this SAP's information.  If there is
    * a problem with the memory allocation, return error status.
    */
   p_sap = (SAP_RECORD*) os_mem_alloc (p_cm->os_handle, sizeof (SAP_RECORD));
   if (p_sap == NULL)
      return STATUS_K_RESOURCES;

   /* Copy the parameters into the SAP record. */

   p_sap->rcv_callback           = rcv_callback;
   p_sap->vc_notify_callback     = vc_notify_callback;
   p_sap->svc_connect_callback   = svc_connect_callback;
   p_sap->link_status_callback   = link_status_callback;
   p_sap->xmt_done_callback      = xmt_done_callback;
   p_sap->p_text                 = p_sap_text;
   p_sap->cm_handle              = cm_handle;
   p_sap->os_handle              = p_cm->os_handle;
   p_sap->svc_handle             = p_cm->svc_handle;
   p_sap->sap_context            = sap_context;

   /* Initialize SAP Statistics. */

   p_sap->xmt_packet_count  = 0;
   p_sap->rcv_packet_count  = 0;
   p_sap->out_call_count    = 0;
   p_sap->accept_call_count = 0;
   p_sap->reject_call_count = 0;

   /* Add the new SAP record to the SAP list. */

   utl_list_add (p_cm->sap_list, p_sap);

   /* Copy the new handle value into the output variable for the caller. */

   *p_sap_handle = (HANDLE) p_sap;

   EVENT (EM_EVENT, ("New SAP Client registered - '%s'.\n", p_sap_text));

   return STATUS_K_SUCCESS;
   }


static BOOLEAN cm_close_by_sap (HANDLE conn_handle,
                                HANDLE sap_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   RELEASE_INFO    rel_info;

   p_sap = (SAP_RECORD    *) sap_handle;
   p_cm  = (CM_INSTANCE   *) p_sap->cm_handle;
   p_vc  = (MAP_VC_RECORD *) conn_handle;

   if (p_vc->sap_handle == sap_handle)
      {
      vc_linkage_remove (conn_handle);

      p_vc->vc_flags &= ~VC_FLAG_NOTIFICATION_REQUIRED;
      if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN)
         {
         VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);
         p_cm->drv_cm_vc_teardown (p_cm->drv_handle,
                                   VPI_GET (p_vc->vc_id),
                                   VCI_GET (p_vc->vc_id),
                                   p_vc->xmt_pool_context,
                                   p_vc->rcv_pool_context);
         }

      if ((VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPENING) ||
          (VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPEN))
         {
         VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);
         svc_release_info_make (CAUSE_NORMAL, &rel_info);
         svc_cm_rel_req (p_cm->svc_handle,
                         p_vc->call_user_id,
                         p_vc->call_provider_id,
                         &rel_info);
         }
      }

   return TRUE;
   }

/*++
* =====================
* = cm_sap_unregister =
* =====================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_sap_unregister (HANDLE sap_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;

   /* Get structure pointers. */

   p_sap = (SAP_RECORD *)  sap_handle;
   p_cm  = (CM_INSTANCE *) p_sap->cm_handle;

   /* Remove the SAP record from the SAP list. */

   utl_list_delete (p_cm->sap_list, p_sap);

   /* Traverse the VC list and remove all VCs associated with this SAP. */

   map_list_traverse (p_cm->vc_list, cm_close_by_sap, sap_handle);

   /* Deallocate the SAP record. */

   os_mem_dealloc (p_cm->os_handle, p_sap);
   }


/*++
* ====================
* = cm_sap_pvc_setup =
* ====================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_pvc_setup (HANDLE   sap_handle,
                         HANDLE   conn_context,
                         UINT8    vpi,
                         UINT16   vci,
                         QOS     *p_requested_qos,
                         HANDLE  *p_conn_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   STATUS          status;
   UINT32          vc_id;
   HANDLE          unused_handle;
   HANDLE          xmt_pool_context;
   HANDLE          rcv_pool_context;

   p_sap = (SAP_RECORD  *) sap_handle;
   p_cm  = (CM_INSTANCE *) p_sap->cm_handle;

   /* Search mapping to see if this vc_id is in use.  If so, return error
    * status.
    */
   vc_id = VC_ID_BUILD (vpi, vci);
   status = map_key_search (p_cm->vc_list, &vc_id, &unused_handle);

   if (status == STATUS_K_SUCCESS)
      return STATUS_K_BUSY;

   /* Check to see if there are sufficient ATM bandwidth resources to meet
    * the QoS requirements.  If not, return error status.
    */
   status = cm_qos_check (p_sap->cm_handle, *p_requested_qos);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_ATM_RESOURCES;

   /* Assign transmit and receive queues to the new VC.  */
   cm_pool_assign (sap_handle,
                   p_requested_qos,
                   &xmt_pool_context,
                   &rcv_pool_context);

   /* Create a new VC record for this PVC. */

   p_vc = (MAP_VC_RECORD *) os_mem_alloc (p_cm->os_handle,
                                          sizeof (MAP_VC_RECORD));
   if (p_vc == NULL)
      return STATUS_K_RESOURCES;

   /* Fill in the appropriate fields of the VC record. */

   p_vc->vc_id             = vc_id;
   p_vc->xmt_pool_context  = xmt_pool_context;
   p_vc->rcv_pool_context  = rcv_pool_context;
   p_vc->sap_handle        = sap_handle;
   p_vc->cm_handle         = p_sap->cm_handle;
   p_vc->conn_context      = conn_context;
   p_vc->atm_link          = NULL;
   p_vc->xmt_packet_count  = 0;
   p_vc->rcv_packet_count  = 0;
   p_vc->idle_time         = 0;
   p_vc->age_limit         = 0;

   p_vc->vc_flags = VC_FLAGS_INIT |
                    VC_FLAG_PVC   |
                    VC_FLAG_NOTIFICATION_REQUIRED;
   VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_OPENING);

   QOS_COPY (*p_requested_qos, p_vc->requested_qos);

   /* Add the VC record to the VC list. */

   map_element_insert (p_cm->vc_list, (HANDLE) p_vc);

   *p_conn_handle = (HANDLE) p_vc;

   if (p_cm->drv_handle != NULL)
      {
      p_cm->drv_cm_vc_setup (p_cm->drv_handle,
                             vpi,
                             vci,
                             xmt_pool_context,
                             rcv_pool_context,
                             p_requested_qos);

      /* Check to see if the phy status has been changed to PHY_OPEN.  It will
       * be if the drv_cm_vc_setup call completed synchronously.  In this case,
       * return STATUS_K_SUCCESS so we don't confuse the SAP client (It has
       * already been notified of the readiness of the VC.
       */
      if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN)
         {
         EVENT (EM_EVENT, ("PVC Opened, vpi=%d, vci=%d, sap=%s.\n",
                           vpi, vci, p_sap->p_text));
         return STATUS_K_SUCCESS;
         }
      else
         return STATUS_K_PENDING;
      }
   else
      {
      /* This is the drv_handle == NULL case. */

      assert (FALSE);
      }

   return STATUS_K_SUCCESS;
   }


/*++
* =======================
* = cm_sap_pvc_teardown =
* =======================
*
* Status:      IMPLEMENTATION COMPLETE.
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_sap_pvc_teardown (HANDLE   conn_handle)
   {
   MAP_VC_RECORD  *p_vc;
   SAP_RECORD     *p_sap;
   CM_INSTANCE    *p_cm;

   /* Convert the needed handles into structure pointers. */

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD *)    p_vc->sap_handle;
   p_cm  = (CM_INSTANCE *)   p_sap->cm_handle;

   /* Remove any linkage in the mapping that involves this connection. */

   vc_linkage_remove (conn_handle);

   /* Set the connection's vc_status to "Closing". */

   VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);

   /* Request that the lower device driver close the connection. */

   p_cm->drv_cm_vc_teardown (p_cm->drv_handle,
                             VPI_GET (p_vc->vc_id),
                             VCI_GET (p_vc->vc_id),
                             p_vc->xmt_pool_context,
                             p_vc->rcv_pool_context);
   }


/*++
* ====================
* = cm_sap_svc_setup =
* ====================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_svc_setup (HANDLE     sap_handle,
                         HANDLE     conn_context,
                         CONN_INFO *p_conn_info,
                         UINT32     age_limit,
                         HANDLE    *p_conn_handle)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   HANDLE          conn_handle;
   ADDR_ATM        called_addr;
   ADDR_ATM        calling_addr;
   QOS             requested_qos;
   STATUS          status;

   p_sap = (SAP_RECORD  *) sap_handle;
   p_cm  = (CM_INSTANCE *) p_sap->cm_handle;

   /* Copy the called party ATM address from the conn_info block. */

   status = conn_info_called_party_address_copy (p_conn_info, &called_addr);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_INSUFFICIENT_DATA;

   /* Copy the calling party ATM address from the conn_info block. */

   status = conn_info_calling_party_address_copy (p_conn_info, &calling_addr);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_INSUFFICIENT_DATA;

   /* Check to see if there are sufficient ATM bandwidth resources to meet
    * the QoS requirements.  If not, return error status.
    */
   conn_info_qos_copy (p_conn_info, &requested_qos);
   status = cm_qos_check (p_sap->cm_handle, requested_qos);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_ATM_RESOURCES;

   /* Create a record for the new connection which is linked properly to an
    * ATM address record.
    */
   status = vc_record_make (p_sap->cm_handle,
                            TRUE,
                            called_addr,
                            &conn_handle);
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_RESOURCES;

   p_vc = (MAP_VC_RECORD *) conn_handle;

   /* Copy the calling party address into the VC record.  This will be used
    * later to determine which VC is the active VC for a given ATM address.
    */
   ATM_COPY (calling_addr, p_vc->calling_party);

   /* Initialize the fields of the VC record. */

   p_vc->sap_handle    = sap_handle;
   p_vc->conn_context  = conn_context;
   p_vc->age_limit     = age_limit;
   p_vc->vc_flags     |= VC_FLAG_CALLING_PARTY |
                         VC_FLAG_NOTIFICATION_REQUIRED;

   VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_OPENING);

   QOS_COPY (requested_qos, p_vc->requested_qos);

   /* Assign the transmit and receive queues to the VC. */

   cm_pool_assign (sap_handle,
                   &requested_qos,
                   &p_vc->xmt_pool_context,
                   &p_vc->rcv_pool_context);

   /* Call the Signaling module and request the connection setup. */

   status = svc_cm_conn_req (p_cm->svc_handle,
                             p_vc->call_user_id,
                             p_conn_info);

   if (status != STATUS_K_SUCCESS)
      {
      /* If there was a signaling problem, disconnect the VC Record's linkage,
       * remove the record from the vc_list, and deallocate its memory.
       */
      VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_FAIL);
      vc_linkage_update (conn_handle);
      return STATUS_K_SIGNALING_ERROR;
      }

   *p_conn_handle = conn_handle;

   EVENT (EM_EVENT, ("SVC Requested by sap '%s'.\n", p_sap->p_text));

   return STATUS_K_SUCCESS;
   }


/*++
* =======================
* = cm_sap_svc_teardown =
* =======================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_svc_teardown (HANDLE        conn_handle,
                            RELEASE_INFO *p_release_info)
   {
   HANDLE          cm_handle;
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;

   /* Obtain the CM handle associated with this connection. */

   p_vc      = (MAP_VC_RECORD *) conn_handle;
   p_sap     = (SAP_RECORD    *) p_vc->sap_handle;
   cm_handle =                   p_sap->cm_handle;
   p_cm      = (CM_INSTANCE   *) cm_handle;

   /* Start the process of tearing down the SVC. */

   EVENT (EM_EVENT, ("SVC Teardown requested by sap '%s'.\n", p_sap->p_text));

   (void) svc_teardown (conn_handle, cm_handle);

   return STATUS_K_PENDING;
   }

/*++
* ========================
* = cm_sap_svc_add_party =
* ========================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_svc_add_party (HANDLE         conn_handle,
                             ADD_PTY_INFO  *p_add_pty_info)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;

   /* Obtain pointers to the important structures. */

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD    *) p_vc->sap_handle;
   p_cm  = (CM_INSTANCE   *) p_sap->cm_handle;

   /* Call the SVC module and request the add party. */

   (void) svc_cm_add_party_req (p_cm->svc_handle,
                                p_vc->call_user_id,
                                p_vc->call_provider_id,
                                p_add_pty_info);

   return STATUS_K_PENDING;
   }

/*++
* =========================
* = cm_sap_svc_drop_party =
* =========================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_svc_drop_party (HANDLE         conn_handle,
                              DROP_PTY_INFO *p_drop_pty_info)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;

   /* Obtain pointers to the important structures. */

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD    *) p_vc->sap_handle;
   p_cm  = (CM_INSTANCE   *) p_sap->cm_handle;

   /* Call the SVC module and request the drop party. */

   (void) svc_cm_drop_party_req (p_cm->svc_handle,
                                 p_vc->call_user_id,
                                 p_vc->call_provider_id,
                                 p_drop_pty_info);

   return STATUS_K_PENDING;
   }


/*++
* =================
* = cm_sap_xmt_vc =
* =================
*
* Status:   IMPLEMENTATION COMPLETE.
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_xmt_vc (HANDLE    conn_handle,
                      void     *p_data,
                      UINT32    length,
                      UINT32    user_data,
                      AAL_DATA *p_aal_data)
   {
   MAP_VC_RECORD  *p_vc;
   SAP_RECORD     *p_sap;
   CM_INSTANCE    *p_cm;

   /* Convert the needed handles into structure pointers. */

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD    *) p_vc->sap_handle;
   p_cm  = (CM_INSTANCE   *) p_sap->cm_handle;

   if (p_aal_data == NULL)
      p_aal_data = &p_cm->default_aal_data;

   /* Increment the transmit packet counters for the SAP and the VC. */

   p_sap->xmt_packet_count += 1;
   p_vc->xmt_packet_count  += 1;
   p_vc->idle_time          = 0;

   /* Pass the Packet to the lower device driver for Transmission. */

   return p_cm->drv_cm_xmt (p_cm->drv_handle,
                            conn_handle,
                            p_data,
                            length,
                            VPI_GET (p_vc->vc_id),
                            VCI_GET (p_vc->vc_id),
                            p_vc->xmt_pool_context,
                            user_data,
                            p_aal_data);
   }


/*++
* ==================
* = cm_sap_xmt_atm =
* ==================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
STATUS cm_sap_xmt_atm (HANDLE    sap_handle,
                       ADDR_ATM  addr,
                       void     *p_data,
                       UINT32    length,
                       UINT32    user_data,
                       AAL_DATA *p_aal_data)
   {
   STATUS          status;
   SAP_RECORD     *p_sap;
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;
   MAP_ATM_RECORD *p_atm;
   HANDLE          atm_handle;

   /* Cast the sap_handle to a copy of a sap_record pointer for later use. */

   p_sap = (SAP_RECORD *)  sap_handle;
   p_cm  = (CM_INSTANCE *) p_sap->cm_handle;

   /* Look up the ATM address in the address/vc mapping. */

   status = map_key_search (p_cm->atm_list,
                            &addr,
                            &atm_handle);

   /* If the ATM address was not found in the mapping, return error status. */
   
   if (status != STATUS_K_SUCCESS)
      return STATUS_K_NOT_FOUND;

   p_atm = (MAP_ATM_RECORD *) atm_handle;

   /* If the ATM address was found but there is no valid VC for the address,
    * return error status.
    */
   if (p_atm->atm_status != ATM_VALID)
      return STATUS_K_NOT_FOUND;

   /* Get the VC record for this ATM address's active VC. */

   p_vc = (MAP_VC_RECORD *) p_atm->active_vc_link;
   assert (p_vc != NULL);

   /* Increment the transmit packet counters for the SAP and the VC. */

   p_sap->xmt_packet_count += 1;
   p_vc->xmt_packet_count  += 1;
   p_vc->idle_time          = 0;

   /* Pass the Packet to the lower device driver for Transmission. */

   return p_cm->drv_cm_xmt (p_cm->drv_handle,
                            (HANDLE) p_vc,
                            p_data,
                            length,
                            VPI_GET (p_vc->vc_id),
                            VCI_GET (p_vc->vc_id),
                            p_vc->xmt_pool_context,
                            user_data,
                            p_aal_data);
   }


/*++
* ==========================
* = cm_sap_link_status_get =
* ==========================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_sap_link_status_get (HANDLE           sap_handle,
                             ATM_LINK_STATE  *p_link_state)
   {
   SAP_RECORD  *p_sap;
   CM_INSTANCE *p_cm;

   p_sap = (SAP_RECORD  *) sap_handle;
   p_cm  = (CM_INSTANCE *) p_sap->cm_handle;

   *p_link_state = p_cm->link_state;
   }

/****************************************************************************
 * Driver Interface Routines
 ****************************************************************************/

/*++
* ============================
* = cm_drv_callback_register =
* ============================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_drv_call_register(HANDLE                cm_handle,
                          DRV_CM_VC_SETUP       drv_cm_vc_setup,
                          DRV_CM_VC_TEARDOWN    drv_cm_vc_teardown,
                          DRV_CM_XMT            drv_cm_xmt)
   {
   CM_INSTANCE *p_cm;

   p_cm = (CM_INSTANCE *) cm_handle;

   p_cm->drv_cm_vc_setup    = drv_cm_vc_setup;
   p_cm->drv_cm_vc_teardown = drv_cm_vc_teardown;
   p_cm->drv_cm_xmt         = drv_cm_xmt;
   }

/*++
* ====================
* = cm_drv_setup_cfm =
* ====================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_drv_setup_cfm (HANDLE  cm_handle,
                       STATUS  status,
                       UINT8   vpi,
                       UINT16  vci,
                       QOS    *p_actual_qos)
   {
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;
   SAP_RECORD     *p_sap;
   UINT32          vc_id;
   HANDLE          conn_handle;
   STATUS          map_status;
   RELEASE_INFO    release_info;

   /* Convert the cm_handle to a pointer to a CM_INSTANCE structure. */

   p_cm = (CM_INSTANCE *) cm_handle;

   EVENT (EM_TRACE, ("cm_drv_setup_cfm entered. vpi=%d, vci=%d\n", vpi, vci));

   /* Convert the vpi/vci to a vc_id to be used as a search key. */

   vc_id = VC_ID_BUILD (vpi, vci);

   /* Search the vc_list for a matching connection record. */

   map_status = map_key_search (p_cm->vc_list, &vc_id, &conn_handle);

   if (map_status == STATUS_K_SUCCESS)
      {
      /* Obtain pointers to the SAP and VC record stuctures. */

      p_vc  = (MAP_VC_RECORD *) conn_handle;
      p_sap = (SAP_RECORD *)    p_vc->sap_handle;

      /* Verify that the phy_status of this record is indeed PHY_PENDING. */

      assert (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPENING);

      if (status == STATUS_K_SUCCESS)
         {
         /* If this confirmation indicates successful VC setup:
          *
          * Copy the actual QoS into the connection record and update its
          * status to VC_VALID.
          */
         QOS_COPY (*p_actual_qos, p_vc->actual_qos);
         VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_OPEN);

         if ((VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPEN) ||
             (p_vc->vc_flags & (VC_FLAG_PVC | VC_FLAG_RVC)))
            {
            /* Update the linkage to and from this VC record in the mapping. */

            vc_linkage_update (conn_handle);

            /* Notify the SAP client that its connection is now ready for
             * use.
             */
            if ((p_sap->vc_notify_callback != NULL) &&
                (p_vc->vc_flags & VC_FLAG_NOTIFICATION_REQUIRED))
               {
               EVENT (EM_EVENT, ("VC Ready, notifying sap '%s'.\n",
                                 p_sap->p_text));
               p_sap->vc_notify_callback (p_vc->conn_context,
                                          conn_handle,
                                          CONN_VC_READY,
                                          NULL,
                                          NULL,
                                          VC_CALLING_PARTY (p_vc->vc_flags));
               p_vc->vc_flags &= ~VC_FLAG_NOTIFICATION_REQUIRED;
               }
            }
         }
      else
         {
         /* If the VC setup resulted in failure, notify the SAP client of
          * the VC release.
          */
         if (p_sap->vc_notify_callback != NULL)
            p_sap->vc_notify_callback (p_vc->conn_context,
                                       conn_handle,
                                       CONN_RELEASE,
                                       (UINT16) CAUSE_RESOURCE_UNAVAIL,
                                       NULL,
                                       VC_CALLING_PARTY (p_vc->vc_flags));

         /* Request the release of this VC. */

         /* Set release cause to RESOURCE_UNAVAILABLE %%%*/

         if ((VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPENING) ||
             (VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPEN))
            svc_cm_rel_req (p_cm->svc_handle,
                            p_vc->call_user_id,
                            p_vc->call_provider_id,
                            &release_info);

         /* Delete the VC record from the VC list and deallocate its memory.
          * Note that there is no VC linkage to disconnect since at this point
          * it has been established that the VC was "opening".
          */
         map_element_delete (p_cm->vc_list, conn_handle);
         os_mem_dealloc (p_cm->os_handle, p_vc);
         }
      }
   else
      {
      /* If no VC record was found matching the vpi/vci, something's wrong.
       * It should never happen that a VC is reported as opened if there is
       * no record that such a VC was requested.
       */
      assert (FALSE);
      }
   }

/*++
* =======================
* = cm_drv_teardown_cfm =
* =======================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_drv_teardown_cfm (HANDLE  cm_handle,
                          UINT8   vpi,
                          UINT16  vci)
   {
   CM_INSTANCE    *p_cm;
   UINT32          vc_id;
   STATUS          status;
   HANDLE          conn_handle;
   MAP_VC_RECORD  *p_vc;

   /* Convert the cm_handle to a pointer to a CM_INSTANCE structure. */

   p_cm = (CM_INSTANCE *) cm_handle;

   /* Convert the vpi/vci to a vc_id to be used as a search key. */

   vc_id = VC_ID_BUILD (vpi, vci);

   /* Search the vc_list for a matching connection record. */

   status = map_key_search (p_cm->vc_list, &vc_id, &conn_handle);

   /* If a matching record is found, delete it from the list and deallocate
    * the memory it is stored in.
    */
   if (status == STATUS_K_SUCCESS)
      {
      vc_linkage_remove (conn_handle);
      map_element_delete (p_cm->vc_list, conn_handle);
      p_vc = (MAP_VC_RECORD *) conn_handle;
      os_mem_dealloc (p_cm->os_handle, p_vc);
      }
   }

/*++
* ==============
* = cm_drv_rcv =
* ==============
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_drv_rcv (HANDLE     cm_handle,
                 UINT32     length,
                 UINT8      vpi,
                 UINT16     vci,
                 UINT32     user_data,
                 AAL_DATA  *p_aal_data,
                 void     **pp_packet)
   {
   CM_INSTANCE    *p_cm;
   UINT32          vc_id;
   STATUS          status;
   HANDLE          conn_handle;
   MAP_VC_RECORD  *p_vc;
   SAP_RECORD     *p_sap;

   /* Convert the cm_handle to a pointer to a CM_INSTANCE structure. */

   p_cm = (CM_INSTANCE *) cm_handle;

   /* Check the link state.  Don't handle the packet if the link is down. */

   if (p_cm->link_state == LINK_DOWN)
      return;

   /* Convert the vpi/vci to a vc_id to be used as a search key. */

   vc_id = VC_ID_BUILD (vpi, vci);

   /* Search the vc_list for a matching connection record. */

   status = map_key_search (p_cm->vc_list, &vc_id, &conn_handle);

   if (status == STATUS_K_SUCCESS)
      {
      /* If a matching connection record is found, obtain pointers to the
       * VC record and SAP record associated with that VC.
       */
      p_vc  = (MAP_VC_RECORD *) conn_handle;
      p_sap = (SAP_RECORD    *) p_vc->sap_handle;

      /* Increment the receive packet counters in the SAP and VC records. */

      p_sap->rcv_packet_count += 1;
      p_vc->rcv_packet_count  += 1;
      p_vc->idle_time          = 0;

      /* Pass the received packet on to the appropriate SAP client. */

      if (p_sap->rcv_callback != NULL)
         p_sap->rcv_callback (p_vc->conn_context,
                              conn_handle,
                              length,
                              user_data,
                              p_aal_data,
                              pp_packet);
      }
   else
      {
      /* If no matching VC record was found, increment the received unknown
       * counter.  This occurs when a packet is received on a VC that has not
       * been opened.
       */
      p_cm->rcv_unknown_count += 1;
      }
   }

/*++
* ===================
* = cm_drv_xmt_done =
* ===================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_drv_xmt_done (HANDLE     cm_handle,
                      HANDLE     conn_handle,
                      void      *p_packet,
                      UINT32     user_data)
   {
   MAP_VC_RECORD  *p_vc;
   SAP_RECORD     *p_sap;

   /* Obtain pointers to the VC and SAP records. */

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD *)    p_vc->sap_handle;

   /* Pass the transmit-done indication up to the SAP client. */

   if (p_sap->xmt_done_callback != NULL)
      p_sap->xmt_done_callback (p_vc->conn_context,
                                p_packet,
                                user_data);
   }

STATUS cm_drv_pool_register (HANDLE         cm_handle,
                             HANDLE         pool_context,
                             POOL_TYPE      pool_type,
                             UINT32         buffer_size_in_bytes,
                             UINT32         depth_in_buffers,
                             UINT32         assignment_mask,
                             HANDLE        *p_pool_handle)
   {
   CM_INSTANCE    *p_cm;
   POOL_RECORD    *p_pool;

   p_cm   = (CM_INSTANCE *) cm_handle;
   p_pool = (POOL_RECORD *) os_mem_alloc (p_cm->os_handle, sizeof (POOL_RECORD));
   if (p_pool == NULL)
      return STATUS_K_RESOURCES;

   p_pool->context         = pool_context;
   p_pool->pool_type       = pool_type;
   p_pool->buffer_size     = buffer_size_in_bytes;
   p_pool->depth           = depth_in_buffers;
   p_pool->assignment_mask = assignment_mask;

   switch (pool_type)
      {
      case POOL_XMT:
         utl_list_add (p_cm->xmt_pool_list, p_pool);  /*%%% Ordered Add Needed */
         break;
      case POOL_RCV:
         utl_list_add (p_cm->rcv_pool_list, p_pool);
         break;
      default:
         EVENT (EM_SERR, ("Invalid Pool Type Registered by Driver.\n"));
         os_mem_dealloc (p_cm->os_handle, p_pool);
         return STATUS_K_ARG_3_BAD;
      }

   *p_pool_handle = (HANDLE) p_pool;
   return STATUS_K_SUCCESS;
   }


STATUS cm_drv_pool_deregister (HANDLE  cm_handle,
                               HANDLE  pool_handle)
   {
   return STATUS_K_SUCCESS;
   }


/****************************************************************************
 * Signaling Interface Routines
 ****************************************************************************/

/* Implementation of the Signaling Protocol in the Connection Manager:
 *
 *  The following shows the calling and called parties in a protocol diagram.
 *  The names at either end of the arrows are those of the routines in the
 *  connection manager that either handle the incoming indications or send
 *  the outgoing requests and responses.  Arrows that originate in the center
 *  of the diagram represent messages that originate in the ATM network.
 *
 *      CALLING PARTY                                   CALLED_PARTY
 *      =============                                   ============
 *                                  SETUP
 *  cm_sap_svc_setup ----------------------------------> cm_svc_conn_ind
 *
 *                      CALL PROCEEDING (optional)
 *  call_proceeding  <-----------------
 *
 *                                 CONNECT
 *  cm_svc_conn_cfm  <---------------------------------- cm_svc_conn_ind
 *
 *                                     CONNECT ACK
 *                                     ----------------> conn_ack
 *
 *                                 RELEASE
 *  cm_sap_svc_teardown -------------------------------> cm_svc_rel_ind
 *
 *                                 RELEASE
 *  cm_svc_rel_ind   <---------------------------------- cm_sap_svc_teardown
 *
 */

/*++
* ===================
* = cm_svc_conn_ind =
* ===================
*
* Status:
*
* Description:
*   Steps that must be taken:
*
*   1.  Get the calling party ATM address from the connection info.  This is
*       an optional element (but is mandatory for LAN Emulation).  If it is
*       present, find or create an ATM address record for the address and
*       link it to the VC record.  If it is not present, leave the VC to ATM
*       address linkage NULL.
*
*   2.  Offer the connection to each qualified SAP client granting it to the
*       first taker.  If there are no takers, or a SAP returns a reason other
*       than "no reason", release the connection and delete the VC record.
*
*   3.  Get the vpi/vci from the connection info.  Assume that the info is
*       present since it is a mandatory element.
*
*   4.  Get the QOS data from the connection info.  Assume that this info is
*       present since it is extracted from mandatory elements.
*
*   5.  Convert the connection info block to a connection status block.
*
*   6.  Call the SVC module and request a connection_response to the network.
*
*   7.  Call the lower device driver and request a VC setup for the connection.
*
*   8.  Set the svc_status and phy_status of the VC to SVC_UNCONFIRMED and
*       PHY_PENDING respectively.
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_conn_ind (HANDLE     cm_handle,
                      CALL_ID    call_provider_id,
                      CONN_INFO *p_conn_info)
   {
   CM_INSTANCE   *p_cm;
   SAP_RECORD    *p_sap;
   MAP_VC_RECORD *p_vc;
   HANDLE         conn_handle;
   STATUS         status;
   QOS            requested_qos;
   ADDR_ATM       calling_addr;
   BOOLEAN        accept;
   UINT32         reject_reason;
   UINT8          vpi;
   UINT16         vci;
   CONN_STATUS    conn_status;
   RELEASE_INFO   rel_info;

   p_cm = (CM_INSTANCE *) cm_handle;

   EVENT (EM_TRACE, ("cm_svc_conn_ind entered.\n"));

   /* Extract the QOS parameters from the conn_info block.  Since these
    * elements are mandatory, we shall assume they are present.
    */
   status = conn_info_qos_copy (p_conn_info, &requested_qos);
   assert (status == STATUS_K_SUCCESS);

   /* Check to see if we have the resources to handle the requested QOS for
    * the VC.  If not, reject the call.
    */
   status = cm_qos_check (cm_handle, requested_qos);
   if (status != STATUS_K_SUCCESS)
      {
      /* Release the connection and exit. */
      
      svc_release_info_make (CAUSE_QOS_CHECK_FAIL, &rel_info);
      svc_cm_rel_req (p_cm->svc_handle,
                      p_cm->next_user_id,
                      call_provider_id,
                      &rel_info);
      p_cm->next_user_id += 1;
      return;
      }

   /* Extract the calling party ATM address from the connection info block.
    * This is an optional element in UNI 3.0 so we can make no assumptions
    * regarding its presence.  The vc_record_make routine will do the right
    * thing for present and absent calling party addresses.
    */
   status = conn_info_calling_party_address_copy (p_conn_info,
                                                  &calling_addr);
   if (status == STATUS_K_SUCCESS)
      status = vc_record_make (cm_handle, TRUE,  calling_addr, &conn_handle);
   else
      status = vc_record_make (cm_handle, FALSE, calling_addr, &conn_handle);

   if (status != STATUS_K_SUCCESS)
      {
      /* Release the connection and exit. */

      svc_release_info_make (CAUSE_RESOURCE_UNAVAIL, &rel_info);
      svc_cm_rel_req (p_cm->svc_handle,
                      p_cm->next_user_id,
                      call_provider_id,
                      &rel_info);
      p_cm->next_user_id += 1;
      return;
      }

   p_vc = (MAP_VC_RECORD *) conn_handle;
   p_vc->call_provider_id   = call_provider_id;

   /* Traverse the list of SAP clients to see if any of them wishes to accept
    * the incoming SVC.  Pre-assign the match and reject_reason variables to
    * ensure that the connection is rejected with the "REJECT_SAP_MISMATCH"
    * reason if there are no elegible SAPs to be offered the connection.  (An
    * eligible SAP is a SAP with a non-NULL svc_connect_callback pointer).
    */
   accept        = FALSE;
   reject_reason = REJECT_SAP_MISMATCH;
   utl_list_traverse (p_cm->sap_list, p_sap)
      {
      if (p_sap->svc_connect_callback != NULL)
         {
         accept = p_sap->svc_connect_callback (p_sap->sap_context,
                                               p_conn_info,
                                               conn_handle,
                                               &p_vc->conn_context,
                                               &p_vc->age_limit,
                                               &reject_reason);

         EVENT (EM_EVENT, ("Sap '%s' %s the connection.\n", p_sap->p_text,
                           accept ? "accepted" : "rejected"))

         if ((accept) || (reject_reason != REJECT_SAP_MISMATCH))
            break;
         }
      }

   if (!accept)
      {
      /* Release the connection using the reject reason. */

      svc_release_info_make (reject_reason, &rel_info);
      svc_cm_rel_req (p_cm->svc_handle,
                      p_vc->call_user_id,
                      p_vc->call_provider_id,
                      &rel_info);

      /* Delete the VC record from the mapping and exit. */

      vc_linkage_remove ((HANDLE) p_vc);
      map_element_delete (p_cm->vc_list, p_vc);
      os_mem_dealloc (p_cm->os_handle, p_vc);
      return;
      }

   /* Put the handle of the accepting SAP client into the VC record.  This
    * completes binding of the SVC to the SAP client.
    */
   p_vc->sap_handle = (HANDLE) p_sap;

   /* Extract the VPI/VCI from the connection info block.  It is assumed that
    * this element is present because it is mandatory.
    */
   status = svc_conn_info_id_get (p_conn_info, &vpi, &vci);
   assert (status == STATUS_K_SUCCESS);
   p_vc->vc_id = VC_ID_BUILD (vpi, vci);

   /* Copy the requested QOS data into the VC record. */

   QOS_COPY (requested_qos, p_vc->requested_qos);

   /* Set the SVC and PHY state values.  The PHY status will become OPEN
    * when the lower device driver confirms the VC setup.
    */
   VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_OPENING);
   VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_OPENING);
   p_vc->vc_flags |= VC_FLAG_NOTIFICATION_REQUIRED | VC_FLAG_VC_ID_PRESENT;

   /* Respond to the network by requesting the transmission of a connection
    * response message.
    */
   svc_conn_info_conn_sts_convert (p_conn_info, &conn_status);
   status = svc_cm_conn_rsp (p_cm->svc_handle,
                             p_vc->call_user_id,
                             p_vc->call_provider_id,
                             &conn_status);
   if (status != STATUS_K_SUCCESS)
      {
      /* Release the connection using the reject reason. */

      svc_release_info_make (CONN_RSP_CALL_FAILED, &rel_info);
      svc_cm_rel_req (p_cm->svc_handle,
                      p_vc->call_user_id,
                      p_vc->call_provider_id,
                      &rel_info);

      /* Delete the VC record from the mapping and exit. */

      vc_linkage_remove (conn_handle);
      map_element_delete (p_cm->vc_list, p_vc);
      os_mem_dealloc (p_cm->os_handle, p_vc);
      return;
      }

   /* Assign transmit and receive queues to the new VC. */

   cm_pool_assign (p_vc->sap_handle,
                   &p_vc->requested_qos,
                   &p_vc->xmt_pool_context,
                   &p_vc->rcv_pool_context);

   /* Request that the lower device driver set up the new VC. */

   p_cm->drv_cm_vc_setup (p_cm->drv_handle,
                          vpi,
                          vci,
                          p_vc->xmt_pool_context,
                          p_vc->rcv_pool_context,
                          &p_vc->requested_qos);
   }


static void call_proceeding (HANDLE       cm_handle,
                             CALL_ID      call_user_id,
                             CALL_ID      call_provider_id,
                             CONN_STATUS *p_conn_status)
   {
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;
   HANDLE          conn_handle;
   STATUS          status;
   UINT8           vpi;
   UINT16          vci;
   RELEASE_INFO    release_info;

   p_cm = (CM_INSTANCE *) cm_handle;

   EVENT (EM_TRACE, ("call_proceeding entered.\n"));

   /* Search the VC List for the connection. */

   status = call_user_id_find (cm_handle, call_user_id, &conn_handle);
   if (status != STATUS_K_SUCCESS)
      {
      svc_release_info_make (RECORD_NOT_FOUND_ON_STATUS_IND, &release_info);
      (void) svc_cm_rel_req (p_cm->svc_handle,
                             call_user_id,
                             call_provider_id,
                             &release_info);
      return;
      }

   p_vc = (MAP_VC_RECORD *) conn_handle;

   /* Extract the VCI and VPI from the connection status block. */

   status = svc_conn_status_id_get (p_conn_status, &vpi, &vci);

   if (status != STATUS_K_SUCCESS)
      {
      /*** Should the vc record be deleted here? ***/
      svc_release_info_make (CONN_ID_MISSING, &release_info);
      (void) svc_cm_rel_req (p_cm->svc_handle,
                             call_user_id,
                             call_provider_id,
                             &release_info);
      return;
      }

   /* Update the VC record with the new info and set its svc_status to
    * UNCONFIRMED.
    */
   p_vc->vc_id     = VC_ID_BUILD (vpi, vci);
   p_vc->vc_flags |= VC_FLAG_VC_ID_PRESENT;
   }

static void conn_ack (HANDLE       cm_handle,
                      CALL_ID      call_user_id,
                      CALL_ID      call_provider_id,
                      CONN_STATUS *p_conn_status)
   {
   CM_INSTANCE    *p_cm;
   MAP_VC_RECORD  *p_vc;
   SAP_RECORD     *p_sap;
   HANDLE          conn_handle;
   STATUS          status;
   RELEASE_INFO    release_info;

   p_cm = (CM_INSTANCE *) cm_handle;

   EVENT (EM_TRACE, ("conn_ack entered.\n"));

   /* Search the VC List for the connection. */

   status = call_user_id_find (cm_handle, call_user_id, &conn_handle);
   if (status != STATUS_K_SUCCESS)
      {
      svc_release_info_make (RECORD_NOT_FOUND_ON_STATUS_IND, &release_info);
      (void) svc_cm_rel_req (p_cm->svc_handle,
                             call_user_id,
                             call_provider_id,
                             &release_info);
      return;
      }

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD *)    p_vc->sap_handle;

   VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_OPEN);

   if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN)
      {
      vc_linkage_update (conn_handle);
      if (p_sap->vc_notify_callback != NULL)
         p_sap->vc_notify_callback (p_vc->conn_context,
                                    (HANDLE) p_vc,
                                    CONN_VC_READY,
                                    NULL,
                                    NULL,
                                    VC_CALLING_PARTY (p_vc->vc_flags));
      }
   }


/*++
* =======================
* = cm_svc_conn_sts_ind =
* =======================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_conn_sts_ind (HANDLE       cm_handle,
                          CALL_ID      call_user_id,
                          CALL_ID      call_provider_id,
                          CONN_STATUS *p_conn_status,
                          EVENT_TYPE   event)
   {
   switch (event)
      {
      case EVENT_CALL_PROCEEDING:

         call_proceeding (cm_handle,
                          call_user_id,
                          call_provider_id,
                          p_conn_status);
         break;

      case EVENT_CONN_ACK:

         conn_ack (cm_handle,
                   call_user_id,
                   call_provider_id,
                   p_conn_status);
         break;

      default:
      
         assert (FALSE);
      }
   }

/*++
* ===================
* = cm_svc_conn_cfm =
* ===================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:   *** Check Flow in this routine! ***
*
--*/
void cm_svc_conn_cfm (HANDLE       cm_handle,
                      CALL_ID      call_user_id,
                      CALL_ID      call_provider_id,
                      CONN_STATUS *p_conn_status)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   HANDLE          conn_handle;
   STATUS          status;
   UINT8           vpi;
   UINT16          vci;
   RELEASE_INFO    release_info;

   p_cm = (CM_INSTANCE *) cm_handle;

   EVENT (EM_TRACE, ("cm_svc_conn_cfm entered.\n"));

   /* Search the VC List for the connection. */

   status = call_user_id_find (cm_handle, call_user_id, &conn_handle);
   if (status != STATUS_K_SUCCESS)
      {
      svc_release_info_make (RECORD_NOT_FOUND_ON_CONN_CFM, &release_info);
      (void) svc_cm_rel_req (p_cm->svc_handle,
                             call_user_id,
                             call_provider_id,
                             &release_info);
      return;
      }

   p_vc = (MAP_VC_RECORD *) conn_handle;
   p_vc->call_provider_id = call_provider_id;

   /* If the vpi/vci of the connection is still not known, extract them from
    * the connection status block.
    */
   if (!(p_vc->vc_flags & VC_FLAG_VC_ID_PRESENT))
      {
      status = svc_conn_status_id_get (p_conn_status, &vpi, &vci);

      if (status != STATUS_K_SUCCESS)
         {
         /*** Should the vc record be deleted here? (yes - 6/5/95) ***/
         svc_release_info_make (CONN_ID_MISSING, &release_info);
         (void) svc_cm_rel_req (p_cm->svc_handle,
                                call_user_id,
                                call_provider_id,
                                &release_info);
         return;
         }

      /* Put the vpi/vci into the VC record. */

      p_vc->vc_id = VC_ID_BUILD (vpi, vci);

      /* Verify that the phy status is PHY_NONE.  It must be because this is
       * the first time we have learned about the vpi/vci of the connection.
       * We know this because the svc_status was SVC_PENDING which means we
       * have not yet learned the vpi/vci.
       */
      assert (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_CLOSED);
      }

   /* The connection setup protocol is complete.  Mark the SVC status as
    * OPEN.
    */
   VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_OPEN);

   /* If the device driver already confirmed the physical setup of the VC,
    * it is time to complete the setup.  This can only have happened if we
    * received a call_proceeding from the network that told us the vpi/vci
    * of the connection.
    */
   if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN)
      {
      /* Get a pointer to the connection's SAP handle. */

      p_sap = (SAP_RECORD *) p_vc->sap_handle;

      /* Update the mapping linkage to the newly opened VC. */

      vc_linkage_update (conn_handle);

      /* Notify the SAP client that the VC is ready to be used. */

      if (p_sap->vc_notify_callback != NULL)
         p_sap->vc_notify_callback (p_vc->conn_context,
                                    conn_handle,
                                    CONN_VC_READY,
                                    NULL,
                                    NULL,
                                    VC_CALLING_PARTY (p_vc->vc_flags));
      }
   else if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_CLOSED)
      {
      VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_OPENING);
      p_cm->drv_cm_vc_setup (p_cm->drv_handle,
                             VPI_GET (p_vc->vc_id),
                             VCI_GET (p_vc->vc_id),
                             p_vc->xmt_pool_context,
                             p_vc->rcv_pool_context,
                             &p_vc->requested_qos);
      }

   return;
   }

/*++
* ==================
* = cm_svc_rel_ind =
* ==================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*   There is a mutual exclusion issue here.  I have implemented the code
*   based on an internal semaphore that protects two small critical sections.
*   This needs to be understood by the team and supported or changed.
*
--*/
void cm_svc_rel_ind (HANDLE        cm_handle,
                     CALL_ID       call_user_id,
                     CALL_ID       call_provider_id,
                     RELEASE_INFO *p_rel_info)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   HANDLE          conn_handle;
   BOOLEAN         svc_flag;
   BOOLEAN         phy_flag;
   STATUS          status;
   RELEASE_INFO    release_info;

   p_cm = (CM_INSTANCE *) cm_handle;

   /* Search the VC List for the connection. */

   status = call_user_id_find (cm_handle, call_user_id, &conn_handle);
   if (status != STATUS_K_SUCCESS)
      {
      svc_release_info_make (RECORD_NOT_FOUND_ON_REL_IND, &release_info);
      (void) svc_cm_rel_req (p_cm->svc_handle,
                             call_user_id,
                             call_provider_id,
                             &release_info);
      return;
      }

   p_vc = (MAP_VC_RECORD *) conn_handle;

   /* Initialize the SVC and PHY flags. */

   svc_flag = FALSE;
   phy_flag = FALSE;

   /* Perform the critical test and set operations to determine if we are the
    * first to transition this VC from OPEN to CLOSING.
    */
   /* CRITICAL_SECTION_ENTER */
   if ((VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPENING) ||
       (VC_SVC_STATE (p_vc->vc_flags) == VC_STATE_OPEN))
      {
      VC_SVC_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);
      svc_flag         = TRUE;
      }
   if (VC_PHY_STATE (p_vc->vc_flags) == VC_STATE_OPEN)
      {
      VC_PHY_STATE_SET (p_vc->vc_flags, VC_STATE_CLOSING);
      phy_flag         = TRUE;
      }
   /* CRITICAL_SECTION_EXIT */

   /* If we transitioned this VC from SVC_OPEN to SVC_CLOSING, notify the
    * SAP client of the release of the connection.
    */
   vc_linkage_update (conn_handle);
   if (svc_flag)
      {
      p_sap = (SAP_RECORD *) p_vc->sap_handle;
      if (p_sap->vc_notify_callback != NULL)
         p_sap->vc_notify_callback (p_vc->conn_context,
                                    (HANDLE) p_vc,
                                    CONN_RELEASE,
                                    CAUSE_GET (p_rel_info),
                                    NULL,
                                    VC_CALLING_PARTY (p_vc->vc_flags));
      }

   /* If we transitioned this VC from PHY_OPEN to PHY_CLOSING, request that
    * the lower device driver close the VC.
    */
   if (phy_flag)
      {
      p_cm->drv_cm_vc_teardown (p_cm->drv_handle,
                                VPI_GET (p_vc->vc_id),
                                VCI_GET (p_vc->vc_id),
                                p_vc->xmt_pool_context,
                                p_vc->rcv_pool_context);
      }

   return;
   }

/*++
* ======================
* = cm_svc_restart_ind =
* ======================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_restart_ind (HANDLE      cm_handle,
                         RST_REASON  reason)
   {
   CM_INSTANCE    *p_cm;

   p_cm = (CM_INSTANCE *) cm_handle;

   map_list_traverse (p_cm->vc_list, &svc_teardown, cm_handle);

   return;
   }

/*++
* ========================
* = cm_svc_add_party_ind =
* ========================
*
* Status:      NOT IMPLEMENTED
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_add_party_ind (HANDLE        cm_handle,
                           CALL_ID       call_user_id,
                           CALL_ID       call_provider_id,
                           ADD_PTY_INFO *p_add_pty_info)
   {
   }

/*++
* ========================
* = cm_svc_add_party_cfm =
* ========================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_add_party_cfm (HANDLE          cm_handle,
                           CALL_ID         call_user_id,
                           CALL_ID         call_provider_id,
                           ADD_PTY_STATUS *p_add_pty_status)
   {
   CM_INSTANCE    *p_cm;
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   HANDLE          conn_handle;
   STATUS          status;
   RELEASE_INFO    release_info;

   p_cm = (CM_INSTANCE *) cm_handle;

   status = call_user_id_find (cm_handle, call_user_id, &conn_handle);
   if (status != STATUS_K_SUCCESS)
      {
      svc_release_info_make (RECORD_NOT_FOUND_ON_REL_IND, &release_info);
      (void) svc_cm_rel_req (p_cm->svc_handle,
                             call_user_id,
                             call_provider_id,
                             &release_info);
      return;
      }

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD *)    p_vc->sap_handle;

   if (p_sap->vc_notify_callback != NULL)
      p_sap->vc_notify_callback (p_vc->conn_context,
                                 conn_handle,
                                 CONN_PARTY_ADD,
                                 NULL,
                                 p_add_pty_status->endptRef.endptRefVal.val,
                                 VC_CALLING_PARTY (p_vc->vc_flags));
   }

/*++
* ============================
* = cm_svc_add_party_rej_ind =
* ============================
*
* Status:      IMPLEMENTATION COMPLETE
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_add_party_rej_ind (HANDLE         cm_handle,
                               CALL_ID        call_user_id,
                               CALL_ID        call_provider_id,
                               DROP_PTY_INFO *p_drop_pty_info)
   {
   SAP_RECORD     *p_sap;
   MAP_VC_RECORD  *p_vc;
   HANDLE          conn_handle;
   STATUS          status;

   status = call_user_id_find (cm_handle, call_user_id, &conn_handle);
   if (status != STATUS_K_SUCCESS)
      return;

   p_vc  = (MAP_VC_RECORD *) conn_handle;
   p_sap = (SAP_RECORD *)    p_vc->sap_handle;

   if (p_sap->vc_notify_callback != NULL)
      p_sap->vc_notify_callback (p_vc->conn_context,
                                 conn_handle,
                                 CONN_PARTY_DROP,
                                 CAUSE_GET        (p_drop_pty_info),
                                 p_drop_pty_info->endptRef.endptRefVal.val,
                                 VC_CALLING_PARTY (p_vc->vc_flags));
   }

/*++
* =========================
* = cm_svc_drop_party_ind =
* =========================
*
* Status:      NOT IMPLEMENTED
*
* Description:
*
* Testing:
*
* Issues:
*
--*/
void cm_svc_drop_party_ind (HANDLE         cm_handle,
                            CALL_ID        call_user_id,
                            CALL_ID        call_provider_id,
                            DROP_PTY_INFO *p_add_pty_info)
   {
   }

void conn_vcid_get (HANDLE  conn_handle,
                    UINT8  *p_vpi,
                    UINT16 *p_vci)
   {
   MAP_VC_RECORD *p_vc;

   p_vc = (MAP_VC_RECORD *) conn_handle;

   if (p_vc == NULL)
      return;

   *p_vpi = VPI_GET (p_vc->vc_id);
   *p_vci = VCI_GET (p_vc->vc_id);
   }

