/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*++
* Module Name:
*   line_up.h
*
* Abstract:
*   This file is the interface to the line-up functionality.  Line-up is when
*   both ends of a link agree to the values to use on the link.
*
* Authors:
*   DMW - Douglas M. Washabaugh
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date       Name  Description
*   10-Feb-95  DMW   Created.
*   15-May-95  TLR   Added Event Reporting Support.
*    3-Aug-95  TLR   Added uni-version call.
*
* Description:
*   The line-up process starts as soon as this module receives a link-up 
*   notification from the ilmi_io module.  Line-up uses ILMI to read remote MIB
*   attributes (DEC MIB and ATM MIB) and cache their values.  Basically, 
*   this module is a database that initializes itself when the link comes up.  
*
*   Once this is complete, the connection manager is called to notify it that
*   line-up is complete.  The connection manager propagates this line-up 
*   indication to all interested parties.
*
*   When the interested parties receive the line-up indication, they
*   can either use the dec_mib or atm_mib routines to read remote objects, 
*   or they can directly call this module.  However, since this module requires
*   SNMP names, the indirect calls to dec_mib and atm_mib are simpler.
*
*   The values which are used on the link depend upon the values on the local
*   side and the remote side of the link.  For instance, Digital FLOWmaster
*   flow-control is used only if both sides can support it.  It is up to 
*   each entity to determine the actual value used on the link.  It is also 
*   the responsibility of the entity to verify the correctness of the values.
*
*   If the link goes down, then no values are available until the line-up
*   has run and completed.
*
* Call Sequence:
*   - line_up_create
*      - line_up_event_mask_set
*      - line_up_get
*      - line_up_destroy
*
* Call Scenario:
*
*   - line_up_create ............... Driver calls this to create an instance.
*
*   - line_up_link_status_change ... Called by ilmi_io to notify that the link
*                                    is now available.
*
*   - <call to connection mgr>   ... The link up process has completed, and now
*                                    line-up calls the connection manager.
*                                    The connection manager notifies all 
*                                    interested parties.
*
*   - line_up_get .................. Interested parties call to get values.
*   -  : : : : : : : : : :  ........ Other values.
*   - line_up_get  ................. Last value.
*
*
*   - line_up_destroy .............. The device/adapter is no longer being used.
*
* Issues:
*   none
--*/

/*++
* ==================
* = line_up_create =
* ==================
*
* Overview:
*   This function creates an instance of this module and configures it.  The
*   line-up module registers with ilmi_io to receive link change notifications.
*   This MUST be the first call made to this module.  The returned handle is
*   required by all calls to this module.
*
* Arguments:
*   os_handle             - (IN)  Handle to use when calling the utl_os module.
*
*   cm_handle             - (IN)  Handle to use when calling the connection mgr.
*
*   ilmi_io_handle        - (IN)  Handle to use when calling ilmi_io.
*
*   p_text                - (IN)  Pointer to a text string that describe this
*                                 instance.  This pointer must remain valid
*                                 until the module is destroyed.
*
*   p_line_up_handle      - (OUT) Handle that is required for subsequent calls
*                                 to this module.
*
* Returns:
*   STATUS_K_SUCCESS -   Instance created.
*   STATUS_K_RESOURCES - Insufficient resources, instance not created.
*
* Preconditions:
*   none
*
* Postconditions:
*   If success is returned, *p_line_up_handle is valid.
--*/                                              

STATUS line_up_create (HANDLE  os_handle,
                       HANDLE  agent_handle,
                       HANDLE  cm_handle,
                       HANDLE  ilmi_io_handle,
                       char   *p_text,                      
                       HANDLE *p_line_up_handle); 

/*++
* ===================
* = line_up_destroy =
* ===================
*
* Overview:
*   This function destroys the instance of this module.  This must be the last
*   call made to the instance.
*
* Arguments:
*   line_up_handle - (IN) The handle returned by line_up_create.
*
* Returns:
*   none
*
* Preconditions:
*   The ilmi_io module is destroyed first because it has a callback to this
*   module.  If the destroy order is incorrect, then a callback to a 
*   non-existent routine would be made.
*
* Postconditions:
*   No more calls to this instance can be made.
--*/                                              

void line_up_destroy (HANDLE line_up_handle); 


/*++
* ==========================
* = line_up_event_mask_set =
* ==========================
*
* Overview:
*   Sets the event mask that defines which events are reported from this
*   module.
*
* Arguments:
*   line_up_handle - (IN) The handle returned by line_up_c.reate.
*   event_mask     - (IN) The new value of the event mask;
*
* Returns:
*   none
*
* Preconditions:
*    The line_up_handle is valid.
*
* Postconditions:
*   none
--*/                                              

void line_up_event_mask_set (HANDLE line_up_handle,
                             UINT32 event_mask); 


/*++
* ===================
* = line_up_obj_get =
* ===================
*
* Overview:
*   This function gets the value of the specified object ID.  Typically this
*   is only called by one of the MIBs.  If a line-up value is needed, use
*   one of the MIB routines to retrieve it.
*
*   If the status code STATUS_K_NOT_FOUND is returned, and the requested object
*   was in the DEC MIB, then this indicates that the remote side does not
*   support the DEC MIB.
*
* Arguments:
*   line_up_handle       - (IN)  The handle returned by line_up_create.
*
*   p_obj_id             - (IN)  Pointer to the complete encoded name (including
*                                the instance portion) of the object.
*
*   obj_id_len           - (IN)  The number of bytes in the buffer pointed to
*                                p_obj_id.
*
*   buff_len             - (IN)  The size of the buffer pointed to by p_buff.
*
*   p_buff               - (OUT) Pointer to the where the object's value is
*                                written.  The buffer should be large enough to
*                                hold the requested data.
*
*   p_bytes_written      - (OUT) The number of bytes that were written to the 
*                                buffer.  If success is not returned, then this
*                                indicates the required size of the buffer.
* Returns:
*   STATUS_K_SUCCESS           - Refer to postconditions.
*   STATUS_K_NOT_FOUND         - Object name was not known. 
*   STATUS_K_FAILURE           - Stored data was corrupt.
*   STATUS_K_BUFFER_TOO_SMALL  - Output buffer is too small for the value.
*   STATUS_K_NOT_READY         - Line-up has not yet completed.
*
* Preconditions:
*   none
*
* Postconditions:
*   The object's value was written to the buffer.
--*/                                              

STATUS line_up_obj_get (HANDLE  line_up_handle,
                        UINT8  *p_obj_id,
                        UINT16  obj_id_len,
                        UINT16  buff_len,
                        void   *p_buff,
                        UINT16 *p_bytes_written); 

/*++
* ===========================
* = line_up_uni_version_get =
* ===========================
*
* Overview:
*   This function returns the UNI version that is being used on the physical
*   port.
*
* Arguments:
*   line_up_handle  - (IN)  The handle returned by line_up_create.
*   p_uni_version   - (OUT) The version number of the UNI supported.
*                           (see defines below)
*
* Returns:
*   STATUS_K_SUCCESS           - Refer to postconditions.
*   STATUS_K_FAILURE           - UNI Version is unknown (not 3.0).
*   STATUS_K_NOT_READY         - Line-up has not yet completed.
*
* Preconditions:
*   none
*
* Postconditions:
*   The uni version is returned.
--*/                                              
#define UNI_3_0   0x30
#define UNI_3_1   0x31
#define UNI_4_0   0x40

STATUS line_up_uni_version_get (HANDLE  line_up_handle,
                                UINT8  *p_uni_version);

