/*************************************************************************
  (c) Copyright.  Digital Equipment Corporation, 1995.  All Rights
  Reserved.

  Permission is hereby granted to use, copy, modify, or enhance this 
  software freely, as long as the foregoing copyright of Digital Equipment
  Corporation and this notice are retained on the software.  This 
  software may not be distributed or sublicensed for a fee.  Digital      
  makes this software available "AS IS" and without warranties of any
  kind.  
 *************************************************************************/

/*
* Module Name:
*   proxy.h
*   
* Overview:
*   This file describes the interface between the LAN Emulation Client
*   modules and an edge device fast path.  This interface is only used
*   if the LEC is acting as a proxy.  If the symbol PROXY is not defined,
*   this include file will never be invoked.
*
*   This interface describes functions that are implemented in the fast path
*   and called from the LAN Emulation Client module.
*
* Authors:
*   TLR - Theodore L. Ross
*
* Modification History:
*   Date       Name  Description 
*   31-Jul-95  TLR   Created.
*
* Description:
*   ATM edge devices (bridges and routers) are often implemented with hardware
*   fast paths.  This interface describes the hooks provided in the common
*   LEC that allow the LEC to support the fast path function.  These functions
*   are only relevant when the LEC is acting as a proxy.
*
*   The proxy interface tells the fast path about new MAC-address-to-VCC
*   mappings when they become usable.  It also informs the fast path of
*   the LEC instance's multicast send and received VCCs in case it wants
*   to handle broadcasts and multicasts its own way.
*/

/*************************************************************************
 * Types that are specific to this interface.                            *
 *************************************************************************/

#define PROXY_FORMAT_802_3  8023
#define PROXY_FORMAT_802_5  8025

/*************************************************************************
 * Function Prototypes for this interface.                               *
 *************************************************************************/

/*++
* =========================
* = proxy_data_direct_add =
* =========================
*
* Overview:
*   This routine is invoked when a new data direct VCC is opened.  It allows
*   the edge-device's fast path to be programmed with a direct MAC address to
*   VCC mapping.  Once this mapping is in place, the LEC module is no longer
*   used to carry data traffic for that VCC.
*
* Arguments:
*   lport_handle  (IN) - Logical Port handle provided at ELAN registration.
*   esi           (IN) - The MAC address associated with the MAC-VC mapping.
*   vpi           (IN) - Virtual Path Identifier of the data direct VCC.
*   vci           (IN) - Virtual Channel Identifier of the data direct VCC.
*   packet_format (IN) - Packet format used on this VCC (choices above).
*   conn_handle   (IN) - Connection Handle of the data direct VCC.
*
* Returns:
*   None
*
* Preconditions:
*   None - Note that the LEC can and probably will report multiple mappings to
*          the same VCC.  This will occur when those MAC addresses are behind
*          another edge device on the ATM network.
*
* Postconditions:
*   None
--*/
void proxy_data_direct_add (HANDLE   lport_handle,
                            ESI      esi,
                            UINT8    vpi,
                            UINT16   vci,
                            UINT16   packet_format,
                            HANDLE   conn_handle);

/*++
* ========================
* = proxy_mcast_send_add =
* ========================
*
* Overview:
*   This routine is invoked when a new multicast send VCC is opened.  This VCC
*   may be used by the fast path to send multicast and broadcast packets on
*   the emulated LAN.
*
* Arguments:
*   lport_handle  (IN) - Logical Port handle provided at ELAN registration.
*   vpi           (IN) - Virtual Path Identifier of the multicast send VCC.
*   vci           (IN) - Virtual Channel Identifier of the multicast send VCC.
*   lec_id        (IN) - LEC_ID to be prepended to packets sent on this VCC.
*                        (Also to be used for echo suppression for any frames
*                        received on this VCC).
*   packet_format (IN) - Packet format used on this VCC (choices above).
*   conn_handle   (IN) - Connection handle of this VCC.  To be used in the
*                        event that a LAN Emulation control frame is received
*                        on this VCC.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   The specified vpi/vci shall be used for all subsequent multicast and
*   broadcast transmissions to this emulated LAN.
--*/
void proxy_mcast_send_add (HANDLE    lport_handle,
                           UINT8     vpi,
                           UINT16    vci,
                           UINT16    lec_id,
                           UINT16    packet_format,
                           HANDLE    conn_handle);


/*++
* =======================
* = proxy_mcast_rcv_add =
* =======================
*
* Overview:
*   This routine is invoked when a new multicast receive VCC is opened.  This
*   VCC will be the one over which multicast, broadcast, and unknown frames
*   will be received.  This information lets the fast path know which logical
*   port the frames come from and provides the LEC-ID for echo suppression.
*
* Arguments:
*   lport_handle  (IN) - Logical Port handle provided at ELAN registration.
*   vpi           (IN) - Virtual Path Identifier of the multicast send VCC.
*   vci           (IN) - Virtual Channel Identifier of the multicast send VCC.
*   lec_id        (IN) - LEC_ID to be used for echo suppression.
*   packet_format (IN) - Packet format used on this VCC (choices above).
*   conn_handle   (IN) - Connection handle of this VCC.  To be used in the
*                        event that a LAN Emulation control frame is received
*                        on this VCC.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
void proxy_mcast_rcv_add (HANDLE     lport_handle,
                          UINT8      vpi,
                          UINT16     vci,
                          UINT16     lec_id,
                          UINT16     packet_format,
                          HANDLE     conn_handle);

/*++
* =================
* = proxy_vcc_del =
* =================
*
* Overview:
*   This routine is invoked when a any mapped VCC is torn down (i.e. data
*   direct or multicast send or receive).  It notifies the edge-device's fast
*   path that the VCC is no longer usable.  If the VCC was a data direct, MAC
*   addresses that were mapped to that VCC must now be treated as unknowns.
*
* Arguments:
*   lport_handle  (IN) - Logical Port handle provided at ELAN registration.
*   vpi           (IN) - Virtual Path Identifier of the VCC.
*   vci           (IN) - Virtual Channel Identifier of the VCC.
*
* Returns:
*   None
*
* Preconditions:
*   None
*
* Postconditions:
*   There is no mapping in the fast-path which involves the specified vpi/vci.
--*/
void proxy_vcc_del (HANDLE  lport_handle,
                            UINT8    vpi,
                            UINT16   vci);

/*++
* ========================
* = proxy_mac_addr_found =
* ========================
*
* Overview:
*   The proxy LEC will receive LE-ARP requests from the LES for MAC addresses
*   it cannot resolve.  The proxy LEC will then use this call to determine if
*   the requested MAC address is present in the bridging database.  The LEC
*   needs only to know whether or not the MAC address is behind this proxy.
*
* Arguments:
*   lport_handle  (IN) - Logical Port handle provided at ELAN registration.
*   esi           (IN) - MAC Address to be searched for in the bridge database.
*
* Returns:
*   Boolean value:  TRUE iff the MAC address is behind the proxy (i.e. found
*                   in the bridge data base for some other port on the bridge
*                   function).
*
* Preconditions:
*   None
*
* Postconditions:
*   None
--*/
BOOLEAN proxy_mac_addr_found (HANDLE   lport_handle,
                              ESI      esi);

